# -*- coding: utf-8 -*-
import collections
import datetime
import gocept.jslint
import icemac.addressbook.addressbook
import icemac.addressbook.interfaces
import icemac.addressbook.person
import icemac.addressbook.startup
import icemac.addressbook.utils
import lxml.etree
import os
import plone.testing
import plone.testing.zca
import plone.testing.zodb
import pytest
import pytz
import transaction
import urlparse
import webtest.forms
import z3c.etestbrowser.wsgi
import zope.app.publication.zopepublication
import zope.app.wsgi.testlayer
import zope.component
import zope.component.hooks
import zope.i18n
import zope.testbrowser.browser
import zope.testbrowser.interfaces
import zope.testbrowser.wsgi


CURRENT_CONNECTION = None
ZODBConnection = collections.namedtuple(
    'ZODBConnection', ['connection', 'rootFolder', 'zodb'])


def createZODBConnection(zodbDB):
    """Create an open ZODB connection."""
    connection = zodbDB.open()
    rootFolder = connection.root()[
        zope.app.publication.zopepublication.ZopePublication.root_name]
    return ZODBConnection(connection, rootFolder, zodbDB)


def pyTestEmptyZodbFixture():
    """Create an empty ZODB prepared to create an address book inside.

    Yields the zodb object.

    Prepared to be usable with a ``pytest.yield_fixture()``.

    """
    layer = plone.testing.zodb.EmptyZODB()
    layer.setUp()
    zodb = layer['zodbDB']
    zope.event.notify(zope.processlifetime.DatabaseOpened(zodb))
    transaction.commit()
    yield zodb
    layer.tearDown()


def pyTestAddressBookFixture(zodb, name):
    """Create an address book in the ZODB.

    Expects an empty ZODB as created by pyTestEmptyZodbFixture() and a name
    for the demo storage layer.

    Yields the zodb object.

    Prepared to be usable with a ``pytest.yield_fixture()``.

    """
    for connection in pyTestStackDemoStorage(zodb, name):
        icemac.addressbook.testing.create_addressbook(
            connection.rootFolder, name='ab', title=u'test addressbook')
        transaction.commit()
        yield connection.zodb


def pyTestStackDemoStorage(zodb, name):
    """Put a demo storage on top of ``zodb`` and yield the connection tuple.

    Prepared to be usable with a ``pytest.yield_fixture()``.

    """
    global CURRENT_CONNECTION
    storage = plone.testing.zodb.stackDemoStorage(zodb, name=name)
    connection = icemac.addressbook.testing.createZODBConnection(storage)
    transaction.begin()
    CURRENT_CONNECTION = connection
    yield connection
    storage.close()
    transaction.abort()
    connection.connection.close()
    CURRENT_CONNECTION = None


def site(connection):
    """Get the address book from the connection and set it as site.

    Usable in a `yield_fixture`.
    Resets the site to its previous value at exit.

    """
    addressbook = connection.rootFolder['ab']
    with zope.component.hooks.site(addressbook):
        yield addressbook


# Layer factories


def SecondaryZCMLLayer(
        name, module, package, bases=(), filename="ftesting.zcml"):
    """Factory to create a new ZCML test layer above an existing one."""
    return plone.testing.zca.ZCMLSandbox(
        name="%sZCML" % name, bases=bases, filename=filename,
        module=module, package=package)


# Test cases

class JSLintTest(gocept.jslint.TestCase):
    """Base test class for JS lint tests."""

    jshint_command = os.environ.get('JSHINT_COMMAND', '/bin/false')
    options = (gocept.jslint.TestCase.options + (
               'evil',
               'eqnull',
               'multistr',
               'sub',
               'undef',
               'browser',
               'jquery',
               'devel'
               ))


# List of users those passwords are not equal to the login name:
USERNAME_PASSWORD_MAP = dict(
    mgr='mgrpw',
    globalmgr='globalmgrpw'
)


class Browser(z3c.etestbrowser.wsgi.ExtendedTestBrowser):
    """Enriched test browser."""

    ROOT_URL_WITHOUT_SLASH = 'http://localhost'
    ROOT_URL = 'http://localhost/'
    # The login URL starts with this string:
    LOGIN_BASE_URL = 'http://localhost/ab/@@loginForm.html?camefrom=http'
    SELENIUM_LOGIN_URL = 'http://localhost/selenium-login'

    ADDRESS_BOOK_DEFAULT_URL = 'http://localhost/ab'
    ADDRESS_BOOK_ADD_URL = 'http://localhost/@@addAddressBook.html'
    ADDRESS_BOOK_WELCOME_URL = 'http://localhost/ab/@@welcome.html'
    ADDRESS_BOOK_2_WELCOME_URL = 'http://localhost/AddressBook/@@welcome.html'
    ADDRESS_BOOK_EDIT_URL = 'http://localhost/ab/@@edit-address_book.html'
    ADDRESS_BOOK_DELETE_PERSONS_URL = (
        'http://localhost/ab/@@delete-address_book-content.html')
    ADDRESS_BOOK_DELETE_URL = 'http://localhost/ab/@@delete-address_book.html'

    MASTER_DATA_URL = 'http://localhost/ab/@@masterdata.html'

    ENTITIES_EDIT_URL = 'http://localhost/ab/++attribute++entities'
    ENTITY_PERSON_LIST_FIELDS_URL = (
        'http://localhost/ab/++attribute++entities/'
        'icemac.addressbook.person.Person')
    ENTITY_PERSON_ADD_FIELD_URL = (
        'http://localhost/ab/++attribute++entities/'
        'icemac.addressbook.person.Person/@@addField.html')
    ENTITIY_PERSON_EDIT_FIELD_URL = (
        'http://localhost/ab/++attribute++entities/'
        'icemac.addressbook.person.Person/Field-1')
    ENTITIY_PERSON_DELETE_FIELD_URL = (
        'http://localhost/ab/++attribute++entities/'
        'icemac.addressbook.person.Person/Field-1/@@delete.html')

    KEYWORDS_LIST_URL = 'http://localhost/ab/++attribute++keywords'
    KEYWORD_ADD_URL = (
        'http://localhost/ab/++attribute++keywords/@@addKeyword.html')
    KEYWORD_EDIT_URL = 'http://localhost/ab/++attribute++keywords/Keyword'
    KEYWORD_DELETE_URL = (
        'http://localhost/ab/++attribute++keywords/Keyword/@@delete.html')

    PERSON_ADD_URL = 'http://localhost/ab/@@addPerson.html'
    PERSON_CLONE_URL = 'http://localhost/ab/Person/@@clone.html'
    PERSON_EDIT_URL = 'http://localhost/ab/Person'
    PERSONS_LIST_URL = 'http://localhost/ab/@@person-list.html'
    PERSON_DELETE_URL = 'http://localhost/ab/Person/@@delete_person.html'
    PERSON_DELETE_ENTRY_URL = (
        'http://localhost/ab/Person/@@delete_entry.html')
    PERSON_EXPORT_URL = 'http://localhost/ab/Person/@@export.html'

    POSTAL_ADDRESS_DELETE_URL = (
        'http://localhost/ab/Person/PostalAddress-2/@@delete.html')
    PHONE_NUMBER_DELETE_URL = (
        'http://localhost/ab/Person/PhoneNumber-2/@@delete.html')
    EMAIL_ADDRESS_DELETE_URL = (
        'http://localhost/ab/Person/EMailAddress-2/@@delete.html')
    HOMEPAGE_ADDRESS_DELETE_URL = (
        'http://localhost/ab/Person/HomePageAddress-2/@@delete.html')

    FILE_ADD_URL = 'http://localhost/ab/Person/@@addFile.html'
    FILE_DELETE_URL = 'http://localhost/ab/Person/File/@@delete.html'
    FILE_DOWNLOAD_URL = 'http://localhost/ab/Person/File/download.html'

    PRINCIPALS_LIST_URL = 'http://localhost/ab/++attribute++principals'
    PRINCIPAL_ADD_URL = (
        'http://localhost/ab/++attribute++principals/@@addPrincipal.html')
    PRINCIPAL_EDIT_URL = 'http://localhost/ab/++attribute++principals/2'
    PRINCIPAL_EDIT_URL_1 = 'http://localhost/ab/++attribute++principals/1'
    PRINCIPAL_DELETE_URL = (
        'http://localhost/ab/++attribute++principals/2/@@delete_user.html')
    PRINCIPAL_DELETE_URL_1 = (
        'http://localhost/ab/++attribute++principals/1/@@delete_user.html')

    SEARCH_URL = 'http://localhost/ab/@@search.html'
    SEARCH_BY_NAME_URL = 'http://localhost/ab/@@name_search.html'
    SEARCH_BY_KEYWORD_URL = 'http://localhost/ab/@@multi_keyword.html'
    SEARCH_MULTI_UPDATE_URL = 'http://localhost/ab/@@multi-update'
    SEARCH_MULTI_UPDATE_CHOOSE_FIELD_URL = (
        'http://localhost/ab/multi-update/chooseField')
    SEARCH_MULTI_UPDATE_ENTER_VALUE_URL = (
        'http://localhost/ab/multi-update/enterValue')
    SEARCH_MULTI_UPDATE_CHECK_RESULT_URL = (
        'http://localhost/ab/multi-update/checkResult')
    SEARCH_MULTI_UPDATE_COMPLETE_URL = (
        'http://localhost/ab/@@multi-update-completed')
    SEARCH_DELETE_URL = 'http://localhost/ab/@@delete_persons.html'

    PREFS_URL = 'http://localhost/ab/++preferences++/ab'
    PREFS_TIMEZONE_URL = (
        'http://localhost/ab/++preferences++/ab.timeZone/@@index.html')

    INSPECTOR_VIEW_URL = PERSONS_LIST_URL + '/@@inspector'
    INSPECTOR_OBJECT_URL = KEYWORDS_LIST_URL + '/@@inspector'
    INSPECTOR_ROOT_OBJECT_URL = ROOT_URL + '/@@inspector'

    def login(self, username, password=None):
        """Login a user using basic auth."""
        password = USERNAME_PASSWORD_MAP.get(username, username)
        self.addHeader(
            'Authorization', 'Basic {username}:{password}'.format(
                username=username, password=password))
        return self

    def formlogin(self, username, password, use_current_url=False):
        """Login using the login form."""
        if not use_current_url:
            self.open(self.ADDRESS_BOOK_DEFAULT_URL)
        self.getControl('User Name').value = username
        self.getControl('Password').value = password
        self.getControl('Log in').click()
        assert 'You have been logged-in successfully.' == self.message
        return self

    def lang(self, lang):
        """Set the language for the browser."""
        self.addHeader('Accept-Language', lang)

    def open(self, url):
        super(Browser, self).open(url)
        return self

    @property
    def contents_without_whitespace(self):
        """Browser contents but with removed whitespace."""
        return self.contents.replace(' ', '').replace('\n', '')

    def etree_to_list(self, etree):
        """"Convert an etree into a list (lines without leading whitespace.)"""
        return [x.strip()
                for x in lxml.etree.tostring(etree).split('\n')
                if x.strip()]

    @property
    def message(self):
        """Return the info messages displayed in browser.

        Returns string when there is exactly one message, list otherwise.
        """
        messages = [x.text
                    for x in self.etree.xpath(
                        '//div[@id="info-messages"]/ul/li')]
        if len(messages) == 1:
            return messages[0]
        return messages

    def keyword_search(self, keyword, apply=None):
        """Search for a keyword via keyword-search.

        If `apply` is not `None` select the value from the dropdown and submit
        the search result handler.

        """
        self.open(self.SEARCH_BY_KEYWORD_URL)
        self.getControl('keywords').displayValue = [keyword]
        self.getControl('Search').click()
        if apply:
            self.getControl('Apply on selected persons').displayValue = [apply]
            self.getControl(name='form.buttons.apply').click()

    @property
    def submit_control_names(self):
        """List of the names of the submit controls in the form."""
        return self._get_control_names(
            zope.testbrowser.interfaces.ISubmitControl, self.getForm())

    @property
    def submit_control_names_all_forms(self):
        """List of the names of the submit controls in all forms."""
        forms = [self.getForm(index=index)
                 for index, _ in enumerate(self._getAllResponseForms())]
        names = [
            self._get_control_names(
                zope.testbrowser.interfaces.ISubmitControl, x)
            for x in forms]
        return names

    @property
    def all_control_names(self):
        """List of the names of all controls in the form."""
        return self._get_control_names(
            zope.testbrowser.interfaces.IControl, self.getForm())

    def in_out_widget_select(self, control_name, select_controls):
        """Add a selection to an in-out-widget.

        control_name ... name of the in-out-control, something like
                         'form.widgets.columns'
        select_controls ... list of control instances (item controls of in or
                             out list) which should be selected
        """
        form = self.getForm()
        for control in select_controls:
            webtest_form = form._form
            name = '%s:list' % control_name
            field = webtest.forms.Hidden(
                webtest_form, tag='input', name=name, pos=999,
                value=control.optionValue)
            webtest_form.fields.setdefault(name, []).append(field)
            webtest_form.field_order.append((name, field))

    def html_redirect(self):
        """Redirect as requested by ``<meta http-equiv="refresh" ... />``."""
        from bs4 import BeautifulSoup
        soup = BeautifulSoup(self.contents, "lxml")
        meta = soup.find('meta')
        assert meta is not None, 'No <meta> tag found.'
        assert meta.get('http-equiv') == 'refresh', \
            '<meta http-equiv != "refresh"'
        url = meta.get('content').partition(';url=')[2]
        self.open(url)

    def select_favicon(self):
        """Select the first favicon."""
        icon = self.etree.xpath(
            '//li[@id="form-widgets-favicon-0"]')[0].attrib['data-value']
        self.getControl(name='form.widgets.favicon:list').value = icon

    def _get_control_names(self, interface, form):
        """Get a sorted list of names of controls providing `interface`."""
        return sorted([control.name
                       for control in form.controls
                       if interface.providedBy(control)])


class WebdriverBase(object):  # pragma: no cover (webdriver)
    """Base for all Webdriver classes."""

    def __init__(self, selenium):
        self._selenium = selenium


class Webdriver(WebdriverBase):  # pragma: no cover (webdriver)
    """Wrapper around Selenese."""

    _to_attach = []

    @classmethod
    def attach(cls, factory, attrib_name):
        """Attach a page object to an attribute of the instance."""
        cls._to_attach.append((factory, attrib_name))

    @classmethod
    def detach(cls, factory, attrib_name):
        """Detach a page object from an attribute of the instance."""
        cls._to_attach.remove((factory, attrib_name))

    def __init__(self, selenium):
        super(Webdriver, self).__init__(selenium)
        for factory, attrib_name in self._to_attach:
            setattr(self, attrib_name, factory(selenium))

    def login(self, username, target_path):
        transaction.commit()
        (_, _, path, _, _) = urlparse.urlsplit(Browser.SELENIUM_LOGIN_URL)
        server = self._selenium.server
        password = USERNAME_PASSWORD_MAP.get(username, username)
        url = "http://{username}:{password}@{server}{path}".format(**locals())
        self.open(url)
        self.open(target_path)

    def open(self, path):
        self._selenium.open(path)

    @property
    def message(self):
        return self._selenium.getText('css=#info-messages')

    @property
    def path(self):
        """Return the path of the current URL."""
        return self._selenium.getLocation().replace(
            'http://{}'.format(self._selenium.server), '')

    def windowMaximize(self):
        self._selenium.windowMaximize()


class WebdriverPageObjectBase(WebdriverBase):  # pragma: no cover (webdriver)
    """Base for page object classes to used with to ``Webdriver.attach()``."""

    browser = Browser  # Browser class to get URLs from
    paths = []  # URL attributes on `browser` to be converted to local paths

    def __init__(self, selenium):
        super(WebdriverPageObjectBase, self).__init__(selenium)
        for name in self.paths:
            url = getattr(self.browser, name)
            path = url.replace('http://localhost', '')
            assert getattr(self, name, None) is None, \
                'duplicate name {}'.format(name)
            setattr(self, name, path)


class TimeZoneMixIn(object):  # pragma: no cover (webdriver)
    """Mix-in for page objects to handle time zones."""

    @property
    def timezone(self):
        return self._selenium.getSelectedLabel("id=form-widgets-time_zone")

    @timezone.setter
    def timezone(self, value):
        # Default time zone can be selected:
        self._selenium.select(
            "id=form-widgets-time_zone", "label={}".format(value))


class POAddressBook(WebdriverPageObjectBase, TimeZoneMixIn):
    """Webdriver page object for the address book itself."""

    paths = [
        'ADDRESS_BOOK_DEFAULT_URL',
        'ADDRESS_BOOK_EDIT_URL',
        'ADDRESS_BOOK_2_WELCOME_URL',
        'ROOT_URL',
        'SEARCH_URL',
    ]

    def create(self):
        # On the start page there is a link to add an address book:
        self._selenium.click('link=address book')

    @property
    def title(self):
        return self._selenium.getValue('id=form-widgets-title')

    @title.setter
    def title(self, title):
        self._selenium.clear('id=form-widgets-title')
        self._selenium.type('id=form-widgets-title', title)

    # setter property!
    def startpage(self, value):
        self._selenium.select(
            'id=form-widgets-startpage', 'label={}'.format(value))

    startpage = property(None, startpage)

    def assert_default_favicon_selected(self):
        # Default value of favicon is pre-selected:
        self._selenium.assertCssCount(
            'css=.ui-selected '
            'img[src="/++resource++img/favicon-red-preview.png"]', 1)

    def select_favicon(self, index):
        self._selenium.clickAt(
            'form-widgets-favicon-{}'.format(index), '20,20')

    def assert_favicon_selected(self, index):
        self._selenium.assertCssCount(
            'css=#form-widgets-favicon-{}.ui-selected'.format(index), 1)

    def submit(self, name):
        self._selenium.click('form-buttons-{}'.format(name))


Webdriver.attach(POAddressBook, 'address_book')


class POPreferences(WebdriverPageObjectBase, TimeZoneMixIn):
    """Webdriver page object for the preferences page."""

    paths = [
        'PREFS_URL',
    ]

    def open_time_zone_element(self):
        self._selenium.click("css=fieldset.timeZone")
        self._selenium.waitForElementPresent("id=form-widgets-time_zone")

    def wait_for_fields_visible(self, visible):
        if visible:
            attrib_name = 'waitForVisible'
        else:
            attrib_name = 'waitForNotVisible'
        return getattr(self._selenium, attrib_name)(
            "css=#form-widgets-columns-row")

    def toggle_group(self, css_class):
        self._selenium.click(
            "//fieldset[@class='{}']/legend".format(css_class))

    def select_column(self, label):
        self._selenium.addSelection(
            "id=form-widgets-columns-from", "label={}".format(label))
        self._selenium.click("name=from2toButton")


Webdriver.attach(POPreferences, 'prefs')


# assertion helper functions and helper classes

def assert_forbidden(browser, username, url):
    """Assert accessing a URL is forbidden for a user."""
    browser.login(username)
    with pytest.raises(zope.testbrowser.browser.HTTPError) as err:
        browser.open(url)
    assert 'HTTP Error 403: Forbidden' == str(err.value)


class SiteMenu(object):
    """Helper class to test selections in the site menu."""

    def __init__(self, browser, menu_item_index, menu_item_title,
                 menu_item_URL):
        """Parameters:

        browser ... return value of the ``browser`` fixture.
        menu_item_index ... zero-based index of the position of the item in
                             the menu
        menu_item_title ... title of the menu item
        menu_item_URL ... URL the menu item should point to

        """
        self.browser = browser
        self.menu_item_index = menu_item_index
        self.menu_item_title = menu_item_title
        self.menu_item_URL = menu_item_URL

    def item_selected(self, url):
        self.browser.open(url)
        return bool(
            self.browser.etree.xpath(self._xpath)[0].attrib.get('class'))

    def assert_correct_menu_item_is_tested(self):
        self.browser.open(self.menu_item_URL)
        assert self.menu_item_title == self.browser.etree.xpath(
            '%s/a/span' % self._xpath)[0].text

    @property
    def _xpath(self):
        # xpath is one based!
        return '//ul[@id="main-menu"]/li[%s]' % (self.menu_item_index + 1)


class AddressBookAssertions:
    """Assertions helpful to check automatic installation routines."""

    def __init__(self, address_book):
        self.address_book = address_book

    def has_local_utility(self, iface, name=''):
        """Assert that there is a local utility for the given interface."""
        util = zope.component.queryUtility(
            iface, context=self.address_book, name=name)
        assert util is not None

    def has_attribute(self, attribute, iface, name=''):
        """Assert the address book has an attribute provided as utility."""
        assert iface.providedBy(getattr(self.address_book, attribute))
        self.has_local_utility(iface, name)


class DateTimeClass(object):
    """Helper class to create and format datetime objects."""

    @property
    def now(self):
        return pytz.utc.localize(datetime.datetime.utcnow())

    def format(self, dt, force_date=False):
        """Format a datetime to the format needed in testbrowser."""
        if isinstance(dt, datetime.datetime) and not force_date:
            return dt.strftime('%y/%m/%d %H:%M')
        else:
            return self.format_date(dt)

    def format_date(self, dt):
        return "{0.year} {0.month} {0.day} ".format(dt)

    @staticmethod
    def add(dt, days=0, seconds=0):
        """Add some days and/or seconds to `dt`."""
        return dt + datetime.timedelta(days=days, seconds=seconds)


# Helper functions to create objects in the database

def create_addressbook(parent, name, title=None):
    """Create an address book in `parent`."""
    ab = icemac.addressbook.utils.create_obj(
        icemac.addressbook.addressbook.AddressBook)
    if title is not None:
        ab.title = title
    parent[name] = ab
    return ab


def create(address_book, parent, interface, set_as_default=False, *args, **kw):
    """Create an object using an entity.

    interface ... interface of the object which should be created.
    """
    with zope.component.hooks.site(address_book):
        entity = icemac.addressbook.interfaces.IEntity(interface)
        name = icemac.addressbook.utils.create_and_add(
            parent, entity.getClass(), *args)
        obj = parent[name]
        for key, value in kw.items():
            field = entity.getField(key)
            context = field.interface(obj)
            field.set(context, value)
        if set_as_default:
            field = icemac.addressbook.person.get_default_field(
                entity.interface)
            field.set(parent, obj)
        zope.lifecycleevent.modified(obj)
    return obj


def set_modified(obj, *args):
    """Set the modification date of an object."""
    dt = datetime.datetime(*args, tzinfo=pytz.utc)
    zope.dublincore.interfaces.IZopeDublinCore(obj).modified = dt
    return dt


def delete_field(browser, field_name):
    """Delete a user defined field."""
    manager = browser.login('mgr')
    manager.open(
        browser.ENTITY_PERSON_LIST_FIELDS_URL +
        '/{}/@@delete.html'.format(field_name))
    manager.getControl('Yes').click()
    assert manager.message.endswith('" deleted.')


# Helper functions and classes


class NonCachingPublicationCache(object):
    """Replacement to prevent caching of ZODB connections."""

    def get(self, key):
        return None

    def __setitem__(self, key, value):
        """Do not cache!"""
        pass


class TestHTTPPublicationRequestFactory(
        zope.app.publication.httpfactory.HTTPPublicationRequestFactory):
    """RequestFactory which uses the global `CURRENT_CONNECTION`."""

    def __init__(self):
        super(TestHTTPPublicationRequestFactory, self).__init__(None)
        self._publication_cache = NonCachingPublicationCache()

    @property
    def _db(self):
        return CURRENT_CONNECTION.zodb

    @_db.setter
    def _db(self, value):
        """Ignoring `value` because the global one should always be used."""


def getRootFolder():
    """`TransactionMiddleware` expects a callable to get the root folder."""
    return CURRENT_CONNECTION.rootFolder


def interpolate_insted_of_translate(
        self, msgid, mapping=None, *args, **kw):  # pragma: no cover webdriver
    """Use interpolation instead of translation."""
    return zope.i18n.interpolate(msgid, mapping)


# Helper views

class SeleniumLogin(object):  # pragma: no cover (webdriver)
    """Allow basic auth log-in of global users.

    It prevents rendering an Unauthorized page.
    """

    def __call__(self):
        return 'Selenium login succeeded.'



from icemac.addressbook.i18n import _
import icemac.addressbook.browser.base
import icemac.addressbook.browser.errormessage
import icemac.addressbook.browser.interfaces
import icemac.addressbook.browser.menus.menu
import icemac.addressbook.browser.search.result.handler.update.operators
import icemac.addressbook.browser.wizard
import persistent.mapping
import z3c.wizard.step
import zope.component
import zope.interface
import zope.lifecycleevent


class UpdateWizard(icemac.addressbook.browser.wizard.Wizard):

    label = _(u'Update Wizard')
    confirmationPageName = '@@multi-update-completed'

    def setUpSteps(self):
        return [
            z3c.wizard.step.addStep(self, 'chooseField', weight=1),
            z3c.wizard.step.addStep(self, 'enterValue', weight=2),
            z3c.wizard.step.addStep(self, 'checkResult', weight=3),
        ]


UPDATE_SESSION_KEY = 'search_result_handler:update'


def get_update_data_session(request):
    """Get the session data for the current update."""
    session = icemac.addressbook.browser.base.get_session(request)
    data = session.get(UPDATE_SESSION_KEY, None)
    if data is None:
        data = persistent.mapping.PersistentMapping()
        session[UPDATE_SESSION_KEY] = data
    return data


def clean_update_data_session(request):
    """Clean the session data from the current update."""
    session = icemac.addressbook.browser.base.get_session(request)
    if session.get(UPDATE_SESSION_KEY, None) is not None:
        del session[UPDATE_SESSION_KEY]


@zope.interface.implementer(
    icemac.addressbook.browser.interfaces.IAddressBookBackground)
class SessionStorageStep(icemac.addressbook.browser.wizard.Step):
    """Step which stores its data in a dict in the session."""

    def getContent(self):
        return get_update_data_session(self.request)


def update_persons(persons, entity, field, operator_name, update_value):
    """Update `entity.field` of persons by using function and update_value."""
    errors = dict()
    for person in persons:
        schema_field = icemac.addressbook.entities.get_bound_schema_field(
            person, entity, field)
        current_value = schema_field.get(schema_field.context)
        operator = zope.component.getAdapter(
            current_value,
            icemac.addressbook.browser.search.result.handler.update.operators.
            IOperator, name=operator_name)
        try:
            new_value = operator(update_value)
        except Exception, e:
            errors[person.__name__] = (
                icemac.addressbook.browser.errormessage.render_error(
                    entity, schema_field.__name__, e))
        else:
            try:
                schema_field.set(schema_field.context, new_value)
            except zope.interface.Invalid, e:
                errors[person.__name__] = (
                    icemac.addressbook.browser.errormessage.render_error(
                        entity, schema_field.__name__, e))
            else:
                zope.lifecycleevent.modified(person)
    return errors


def get_fieldname_in_session(fieldname):
    """Return the name of the field used in the session."""
    return 'new_value.%s' % fieldname


update_views = icemac.addressbook.browser.menus.menu.SelectMenuItemOn(
    'multi-update', 'chooseField', 'enterValue', 'checkResult')



from icemac.addressbook.i18n import _
import cStringIO
import datetime
import decimal
import icemac.addressbook.export.base
import icemac.addressbook.interfaces
import xlwt
import zope.i18n
import zope.i18nmessageid
import zope.security.proxy


# fonts
head_font = xlwt.Font()
head_font.name = 'Courier'

group_font = xlwt.Font()
group_font.name = 'Courier'
group_font.bold = True


# styles
default_style = xlwt.XFStyle()

head_style = xlwt.XFStyle()
head_style.font = head_font

group_style = xlwt.XFStyle()
group_style.font = group_font

date_style = xlwt.XFStyle()
date_style.num_format_str = 'DD.MM.YY'

datetime_style = xlwt.XFStyle()
datetime_style.num_format_str = 'DD.MM.YY HH:MM:SS'

value_style_mapping = {
    datetime.date: date_style,
    datetime.datetime: datetime_style,
}


class XLSExport(icemac.addressbook.export.base.BaseExporter):
    """Abstract base class for xls export."""

    file_extension = 'xls'
    mime_type = 'application/vnd.ms-excel'

    def convert_value(self, value):
        """Convert the value for xls export."""
        if value is None:
            return value
        if value.__class__ in (str, unicode, float, int, bool, datetime.date,
                               datetime.datetime, decimal.Decimal):
            return self.translate(value)
        if hasattr(value, '__iter__'):
            return ', '.join(self.convert_value(v) for v in value)
        return self.translate(icemac.addressbook.interfaces.ITitle(value))

    def translate(self, value):
        if not isinstance(value, zope.i18nmessageid.Message):
            return value
        return zope.i18n.translate(value, context=self.request)

    def export(self):
        self.workbook = xlwt.Workbook()
        self.sheet = self.workbook.add_sheet(
            self.translate(_(u'Address book - Export')))
        self.col = 0

        self._export()

        io = cStringIO.StringIO()
        self.workbook.save(io)
        return io.getvalue()

    def write_headlines(self, col, interface, headline):
        self.sheet.write(0, col, headline, group_style)
        fields = icemac.addressbook.interfaces.IEntity(
            interface).getFieldValues()
        for field in fields:
            self.sheet.write(1, col, self.translate(field.title), head_style)
            col += 1
        return col

    def write_cell(self, row, col, value):
        style = value_style_mapping.get(value.__class__, default_style)
        self.sheet.write(row, col, self.convert_value(value), style)

    def write_col(self, interface, obj_getter, headline):
        max_col = self.write_headlines(self.col, interface, headline)
        row = 1
        for person in self.persons:
            row += 1
            obj = obj_getter(person)
            self.write_obj(row, self.col, interface, obj)
        self.col = max_col

    def write_obj(self, row, col, interface, obj):
        idx = 0
        names_fields = icemac.addressbook.interfaces.IEntity(
            interface).getFields()
        for name, field in names_fields:
            # Need to remove the security proxy to access the user
            # defined fields.
            context = field.interface(zope.security.proxy.getObject(obj))
            self.write_cell(row, col + idx, getattr(context, name))
            idx += 1
        return col + idx

    def write_person_data(self):
        self.write_col(
            icemac.addressbook.interfaces.IPerson,
            lambda x: x,
            self.translate(_('person')))

    def get_entities(self):
        return zope.component.getUtility(
            icemac.addressbook.interfaces.IEntities).getMainEntities()


class DefaultsExport(XLSExport):
    """Exports person data and main addresses resp. phone numbers."""

    def _export(self):
        for entity in self.get_entities():
            if entity.interface == icemac.addressbook.interfaces.IPerson:
                self.write_person_data()
            else:
                self.write_col(
                    entity.interface,
                    lambda x: getattr(
                        x, entity.tagged_values['default_attrib']),
                    self.translate(entity.title))


class CompleteExport(XLSExport):
    """Exports person data and all addresses resp. phone numbers."""

    def _export(self):
        for entity in self.get_entities():
            if entity.interface == icemac.addressbook.interfaces.IPerson:
                self.write_person_data()
            else:
                self.write_block(
                    entity.interface, entity.tagged_values['default_attrib'],
                    entity.title)

    def write_block(self, iface, default_attrib, title):
        num_blocks = 0
        blocks_with_header = 0
        start_col = max_col = self.col
        row = 1
        for person in self.persons:
            row += 1
            # write default objs first
            default_obj = getattr(person, default_attrib)
            if default_obj is None:
                continue
            if num_blocks == 0:
                max_col = self.write_headlines(
                    start_col, iface,
                    self.translate(
                        _(u'main ${address}', mapping=dict(address=title))))
                num_blocks = 1
                blocks_with_header = 1
            col = self.write_obj(row, start_col, iface, default_obj)
            # write other objs after default obj
            objs = [obj
                    for obj in icemac.addressbook.utils.iter_by_interface(
                        person, iface)
                    if obj != default_obj]
            if not objs:
                continue
            num_blocks = max(num_blocks, len(objs) + 1)
            while blocks_with_header < num_blocks:
                max_col = self.write_headlines(
                    max_col, iface,
                    self.translate(
                        _(u'other ${address}', mapping=dict(address=title)))
                )
                blocks_with_header += 1
            for obj in objs:
                col = self.write_obj(row, col, iface, obj)
        self.col = max_col



# -*- coding: utf-8 -*-
from icemac.addressbook.principals.interfaces import IPasswordFields
from icemac.addressbook.principals.interfaces import IPrincipal, IRoles
from icemac.addressbook.principals.principals import Principal, created
from icemac.addressbook.utils import create_obj
from zope.interface.verify import verifyObject
import icemac.addressbook.interfaces
import pytest


def test_principals__Principal__1(zcmlS):
    """It fulfills the `IPrincipal` interface."""
    principal = Principal()
    # We need to call the created event handler here, because the person
    # attribute is a descriptor wrapping the object verifyObject expects.
    created(principal, None)
    assert verifyObject(IPrincipal, principal)


def test_principals__Principal__2(zcmlS):
    """It fulfills the `IPasswordFields` interface."""
    principal = Principal()
    assert verifyObject(IPasswordFields, principal)


def test_principals__Principal__3(zcmlS):
    """It fulfills the `IRoles` interface."""
    principal = Principal()
    assert verifyObject(IRoles, principal)


def test_principals__Principal__person__1(address_book, FullPersonFactory):
    """It raises a ValueError if the `person` is set again."""
    principal = create_obj(Principal)
    person = FullPersonFactory(address_book, u'User')
    principal.person = person
    assert principal.person == person
    with pytest.raises(ValueError) as err:
        principal.person = None
    assert ('It is not possible to change the person on a principal.' ==
            str(err.value))


def test_principals__title__1(zcmlS):
    """It returns a default value if the principal has no person assigned."""
    principal = Principal()
    assert '<no person>' == icemac.addressbook.interfaces.ITitle(principal)



import icemac.addressbook.browser.search.result.handler.base


class BaseExport(icemac.addressbook.browser.search.result.handler.base.Base):
    """Base class for exporters.

    Subclasses only have to set the `exporter_class` which should implement
    the interface `icemac.addressbook.export.interfaces.IExporter`.
    """

    exporter_class = NotImplemented

    def __call__(self):
        exporter = self.exporter_class(self.persons, self.request)

        self.request.response.setHeader('Content-Type', exporter.mime_type)
        self.request.response.setHeader(
            'Content-Disposition',
            'attachment; filename=addressbook_export.%s' % (
                exporter.file_extension))

        return exporter.export()



import icemac.addressbook.browser.search.base
import icemac.addressbook.sources
import z3c.menu.ready2go
import z3c.menu.ready2go.item
import z3c.menu.ready2go.manager
import zope.viewlet.manager


class ISearchResultHandlers(z3c.menu.ready2go.ISiteMenu):
    """Collection of search result handlers."""


SearchResultHandlerManager = zope.viewlet.manager.ViewletManager(
    'search-result-handlers', ISearchResultHandlers, bases=(
        z3c.menu.ready2go.manager.MenuManager,))


source = icemac.addressbook.sources.SiteMenuSource(
    icemac.addressbook.browser.search.base.BaseView, SearchResultHandlerManager
)


class SearchResultHandler(z3c.menu.ready2go.item.SiteMenuItem):
    """Specialized menu item to be used in SearchResultHandlerManager."""

    def __eq__(self, other):
        """Each handler has a unique viewName, usable for comparison."""
        if not isinstance(other, SearchResultHandler):
            return False
        return self.viewName == other.viewName



from icemac.addressbook.i18n import _
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
import icemac.addressbook.browser.interfaces
import icemac.addressbook.file.interfaces
import z3c.form.browser.file
import z3c.form.converter
import z3c.form.interfaces
import z3c.form.widget
import zope.component
import zope.schema.interfaces


class FileWidget(z3c.form.browser.file.FileWidget):
    """FileWidget for use with ..file.interfaces.IFile['data']."""

    download_template = ViewPageTemplateFile('widget_download.pt')

    @property
    def href(self):
        view = zope.component.getMultiAdapter(
            (self.context, self.request), name='download.html')
        url = zope.component.getMultiAdapter(
            (view, self.request), name='absolute_url')
        return url

    content = _('Download file')

    @property
    def download(self):
        if icemac.addressbook.file.interfaces.IFile.providedBy(self.context):
            return self.download_template(self)
        return ''


@zope.component.adapter(
    zope.schema.interfaces.IBytes,
    icemac.addressbook.browser.interfaces.IAddressBookLayer)
@zope.interface.implementer(z3c.form.interfaces.IFieldWidget)
def FileFieldWidget(field, request):
    """Adapt IBytes to FileWidget."""
    return z3c.form.widget.FieldWidget(field, FileWidget(request))


class FileUploadDataConverter(z3c.form.converter.FileUploadDataConverter):
    """Coverter which does not read the whole file contents into a variable."""

    def toFieldValue(self, value):
        """Convert value to be stored.

        CAUTION: Does not return the file contents but only the first
        byte, to signal that a file was uploaded.

        """
        if value is None or value == '':
            # When no new file is uploaded, send a signal that we do not want
            # to do anything special.
            return z3c.form.interfaces.NOT_CHANGED

        # We expect here to have a FileUpload instance.  We store the
        # additional FileUpload values on the widget before we loose
        # them.
        self.widget.headers = value.headers
        self.widget.filename = value.filename
        value.seek(0)
        data = value.read(1)
        return data



def test_export__ExportList__1(search_data, browser):
    """..export.ExportList() renders an advice if there are no exporters.

    The actual exporters are tested in their modules.

    """
    browser.login('visitor')
    browser.open(browser.PERSONS_LIST_URL)
    browser.getLink('Hohmuth').click()
    browser.getControl('Export').click()
    assert ('You did not enter enough data of the person, '
            'so no export is possible.' in browser.contents)



from icemac.addressbook.interfaces import IEntities
from zope.testbrowser.browser import HTTPError
import pytest
import zope.component
import zope.component.hooks


"""The sort order of the entities can be changed globally. The places where
   entities are displayed in order are changed accordingly."""


UP_LINK_URL = ('http://localhost/ab/++attribute++entities/'
               'icemac.addressbook.address.PostalAddress/@@up.html')
DOWN_LINK_URL = ('http://localhost/ab/++attribute++entities/'
                 'icemac.addressbook.addressbook.AddressBook/@@down.html')


def get_entity_titles(address_book):
    """Get the entities of the address book ordered by the sort order."""
    with zope.component.hooks.site(address_book):
        entities = zope.component.getUtility(IEntities)
        return [x.title for x in entities.getEntities()]


def test_entities__1(address_book):
    """There is a default sort order of the entities."""
    assert [u'address book', u'person', u'main adresses and numbers',
            u'postal address'] == get_entity_titles(address_book)[:4]


def test_entities__EntitiesTraverser__publishTraverse__1(
        address_book, browser):
    """It raises a HTTP-404 for unknown entity names."""
    browser.login('mgr')
    with pytest.raises(HTTPError) as err:
        browser.open(browser.ENTITIES_EDIT_URL + '/not.existing.entity')
    assert 'HTTP Error 404: Not Found' == str(err.value)


def test_entities__MoveUp__1(address_book, browser):
    """Selecting the `up` link moves the entity one position up in the list."""
    browser.login('mgr')
    browser.open(browser.ENTITIES_EDIT_URL)
    # Let's move the postal address one up:
    assert UP_LINK_URL == browser.getLink('up', index=2).url
    browser.open(UP_LINK_URL)
    assert 'Moved postal address up.' == browser.message
    assert [u'address book', u'person', u'postal address',
            u'main adresses and numbers'] == get_entity_titles(
                address_book)[:4]


def test_entities__MoveDown__1(address_book, browser):
    """Selecting the `down` link moves the entity one position down in list."""
    browser.login('mgr')
    browser.open(browser.ENTITIES_EDIT_URL)
    # Let's move he address book one down:
    assert DOWN_LINK_URL == browser.getLink('down', index=0).url
    browser.open(DOWN_LINK_URL)
    assert 'Moved address book down.' == browser.message
    assert [u'person', u'address book', u'main adresses and numbers',
            u'postal address'] == get_entity_titles(address_book)[:4]


def test_entities__List__1(address_book, browser):
    """The entity sort order is reflected in the entity table."""
    browser.login('mgr')
    browser.open(DOWN_LINK_URL)
    assert ['person', 'address book', 'main adresses and numbers',
            'postal address'] == browser.etree.xpath(
                '//tbody/tr/td[1]/text()')[:4]  # xpath is one-based!


def test_entities__UpLinkColumn__1(address_book, browser):
    """There is no `up` link for the first entity in the table."""
    browser.login('mgr')
    browser.open(browser.ENTITIES_EDIT_URL)
    assert ['down', 'Edit fields'] == browser.etree.xpath(
        '//tbody/tr[1]/td/a/text()')  # xpath is one-based!


def test_entities__DownLinkColumn__1(address_book, browser):
    """There is no `down` link for the last entity in the table."""
    browser.login('mgr')
    browser.open(browser.ENTITIES_EDIT_URL)
    assert ['up', 'Edit fields'] == browser.etree.xpath(
        '//tbody/tr[last()]/td/a/text()')


@pytest.mark.parametrize('username', ('editor', 'visitor'))
def test_entities__security__1(address_book, browser, username):
    """Non-admin users are not able to change the entity sort order."""
    browser.login(username)
    with pytest.raises(HTTPError) as err:
        browser.open(UP_LINK_URL)
    assert 'HTTP Error 403: Forbidden' == str(err.value)
    with pytest.raises(HTTPError) as err:
        browser.open(DOWN_LINK_URL)
    assert 'HTTP Error 403: Forbidden' == str(err.value)



import pytest


def test_session__FlashedSessionCredentialsPlugin__1(address_book, browser):
    """The login screen displays a flash message.

    As cookie authorization is used an unauthorized user is redirected to
    the log-in page.

    """
    browser.open(browser.ADDRESS_BOOK_DEFAULT_URL)
    assert (
        'To log-in enter your username and password and submit the form.' ==
        browser.message)
    assert (
        browser.LOGIN_BASE_URL +
        '%3A%2F%2Flocalhost%2Fab%2F%40%40index.html' == browser.url)


@pytest.mark.parametrize('username, password',
                         (('wrong_username@example.com', '123456789'),
                          ('username@example.com', 'wrong_123456789')))
def test_session__FlashedSessionCredentialsLoginForm__1(
        address_book, UserFactory, browser, username, password):
    """A wrong user name or password at login leads to an error message.

    No time-stamp of this attempt is stored.
    """
    user = UserFactory(address_book, u'Test', u'User', u'username@example.com',
                       u'123456789', ['Visitor'])
    browser.open(browser.ADDRESS_BOOK_DEFAULT_URL)
    browser.getControl('User Name').value = username
    browser.getControl('Password').value = password
    browser.getControl('Log in').click()
    assert ('Login failed. Username and/or password might be wrong.' ==
            browser.message)
    assert 'Please provide Login Information' in browser.contents
    assert (
        browser.LOGIN_BASE_URL +
        '%3A%2F%2Flocalhost%2Fab%2F%40%40index.html' == browser.url)
    assert user.last_login is None
    # User name and password are not re-displayed if there was an error:
    assert '' == browser.getControl('User Name').value
    assert '' == browser.getControl('Password').value


@pytest.mark.parametrize('role', ('Visitor', 'Editor', 'Administrator'))
def test_session__FlashedSessionCredentialsLoginForm__2(
        address_book, UserFactory, DateTime, browser, role):
    """A user is able to log-in using the cookie login form.

    His last log-in timestamp is stored.
    """
    user = UserFactory(address_book, u'Test', u'User', u'username@example.com',
                       u'123456789', [role])
    browser.open(browser.ADDRESS_BOOK_DEFAULT_URL)
    browser.getControl('User Name').value = 'username@example.com'
    browser.getControl('Password').value = '123456789'
    browser.getControl('Log in').click()
    assert 'You have been logged-in successfully.' == browser.message
    assert browser.ADDRESS_BOOK_WELCOME_URL == browser.url
    assert user.last_login is not None
    assert DateTime.add(DateTime.now, seconds=-5) < user.last_login
    browser.getLink('Master data').click()
    browser.getLink('Users').click()
    # Role name and last log-in are shown in user list:
    print browser.contents
    assert (
        '<td>username@example.com</td><td>{}</td><td>{}</td>'.format(
            role, DateTime.format(user.last_login).replace(' ', '')) in
        browser.contents_without_whitespace)


@pytest.mark.parametrize('role', ('Visitor', 'Editor', 'Administrator'))
def test_session__FlashedSessionCredentialsLoginForm__3(
        address_book, UserFactory, browser, role):
    """Selecting the ``Logout``-Link leads to the log-in form."""
    UserFactory(address_book, u'Test', u'User', u'username@example.com',
                u'123456789', [role])
    browser.open(browser.ADDRESS_BOOK_DEFAULT_URL)
    browser.getControl('User Name').value = 'username@example.com'
    browser.getControl('Password').value = '123456789'
    browser.getControl('Log in').click()
    # Logout:
    browser.getLink('Logout').click()
    browser.html_redirect()
    assert (
        ['You have been logged out successfully.',
         'To log-in enter your username and password and submit the form.'] ==
        browser.message)
    assert (
        browser.LOGIN_BASE_URL +
        '%3A%2F%2Flocalhost%2Fab%2F%40%40welcome.html' == browser.url)


def test_session__FlashedSessionCredentialsLoginForm__4(
        address_book, AddressBookFactory, UserFactory, browser):
    """Log-in to another address book not possible.

    It is not possible to log in into another address book besides the one
    the user was created in (password does not get accepted).

    """
    UserFactory(address_book, u'Ad', u'Min', u'admin@example.com',
                u'123456789', ['Administrator'])
    AddressBookFactory('ab2')
    browser.open('http://localhost/ab2')
    assert ('http://localhost/ab2/@@loginForm.html?camefrom='
            'http%3A%2F%2Flocalhost%2Fab2%2F%40%40index.html' == browser.url)
    browser.getControl('User Name').value = 'admin@example.com'
    browser.getControl('Password').value = '123456789'
    browser.getControl('Log in').click()
    assert ('Login failed. Username and/or password might be wrong.' ==
            browser.message)
    assert ('http://localhost/ab2/@@loginForm.html?camefrom='
            'http%3A%2F%2Flocalhost%2Fab2%2F%40%40index.html' == browser.url)



import icemac.addressbook.generations.utils
import zope.component
import zope.pluggableauth.interfaces
import zope.principalannotation.interfaces
import BTrees.OOBTree


@icemac.addressbook.generations.utils.evolve_addressbooks
def evolve(addressbook):
    """Update preferences to new structure."""
    principals = zope.component.getUtility(
        zope.pluggableauth.interfaces.IAuthenticatorPlugin,
        name=u'icemac.addressbook.principals')
    principal_annotations = zope.component.getUtility(
        zope.principalannotation.interfaces.IPrincipalAnnotationUtility)
    for pid in principals.keys():
        prefs = principal_annotations.getAnnotationsById(pid).get(
            zope.preference.preference.pref_key, {})
        if 'personList' not in prefs:
            continue
        for key, value in prefs['personList'].items():
            if key == 'batch_size':
                parent_key = 'ab.personListTab'
            elif key in ['columns', 'order_by', 'sort_direction']:
                parent_key = 'ab.personLists'
            else:
                raise KeyError('Unkown pref key: %s' % key)
            if parent_key not in prefs:
                prefs[parent_key] = BTrees.OOBTree.OOBTree()
            prefs[parent_key][key] = value
        del prefs['personList']



from icemac.addressbook.i18n import _
import icemac.addressbook.metadata.interfaces
import z3c.form.field
import z3c.form.group
import z3c.form.interfaces
import zope.component
import zope.dublincore.interfaces
import zope.preference.interfaces


class MetadataBaseGroup(z3c.form.group.Group):
    """Base class for groups displaying meta data information."""

    label = _('metadata')
    mode = z3c.form.interfaces.DISPLAY_MODE
    fields = NotImplemented

    def updateWidgets(self):
        """See interfaces.IForm"""
        self.widgets = zope.component.getMultiAdapter(
            (self, self.request, self.getContent()),
            z3c.form.interfaces.IWidgets)
        for attrName in ('ignoreRequest', 'ignoreContext', 'ignoreReadonly'):
            value = getattr(self.parentForm.widgets, attrName)
            setattr(self.widgets, attrName, value)
        # we need the value from the group here, not the one from the
        # parent form
        self.widgets.mode = self.mode
        self.widgets.update()


class MetadataGroup(MetadataBaseGroup):
    """Group displaying all meta data information."""

    fields = z3c.form.field.Fields(
        icemac.addressbook.metadata.interfaces.IEditor).select('creator')
    fields += z3c.form.field.Fields(
        zope.dublincore.interfaces.IDCTimes).select('created')
    fields += z3c.form.field.Fields(
        icemac.addressbook.metadata.interfaces.IEditor).select('modifier')
    fields += z3c.form.field.Fields(
        zope.dublincore.interfaces.IDCTimes).select('modified')


def timezone_messageid_factory(message_id):
    """Set currently selected time zone in the mapping of the message id."""
    def factory(ignored):
        prefs = zope.component.getUtility(
            zope.preference.interfaces.IPreferenceGroup, name="ab.timeZone")
        return _(message_id, mapping=dict(timezone=prefs.time_zone))
    return factory


ModifiedLabel = z3c.form.widget.ComputedWidgetAttribute(
    timezone_messageid_factory(_(u'Modification Date (${timezone})')),
    field=zope.dublincore.interfaces.IDCTimes['modified'])

ModifiedHint = z3c.form.widget.StaticWidgetAttribute(
    _(u'The date and time that the object was last modified converted to '
      u'the displayed time zone.'),
    field=zope.dublincore.interfaces.IDCTimes['modified'])

CreatedLabel = z3c.form.widget.ComputedWidgetAttribute(
    timezone_messageid_factory(_(u'Creation Date (${timezone})')),
    field=zope.dublincore.interfaces.IDCTimes['created'])

CreatedHint = z3c.form.widget.StaticWidgetAttribute(
    _(u'The date and time that the object was created converted to '
      u'the displayed time zone.'),
    field=zope.dublincore.interfaces.IDCTimes['created'])

MetadataGroupFieldsNotRequired = z3c.form.widget.StaticWidgetAttribute(
    False,
    view=MetadataGroup)



import icemac.addressbook.conftest
import icemac.addressbook.interfaces
import mock
import pytest
import transaction
import zope.component
import zope.component.hooks


# Fixtures to set-up infrastructure which are usable in tests:

@pytest.yield_fixture(scope='function')
def mailto_data(mailtoDataS):
    """Make a stacked demo storage on search data."""
    mailToConnectionS, persons = mailtoDataS
    for connection in icemac.addressbook.conftest.pyTestStackDemoStorage(
            mailToConnectionS.zodb, 'MailToFunction'):
        yield persons


# Infrastructure fixtures


@pytest.yield_fixture(scope='session')
def mailtoDataS(addressBookS, KeywordFactory, PersonFactory,
                EMailAddressFactory, FullPersonFactory):
    """Create data used in mail-to tests."""
    kws = [u'mail-me']
    for connection in icemac.addressbook.conftest.pyTestStackDemoStorage(
            addressBookS, 'MailToSession'):
        for address_book in icemac.addressbook.conftest.site(connection):
            persons = {}
            # No EMailAddress object
            persons['p0'] = PersonFactory(
                address_book, u'No Mail', keywords=kws)
            persons['p1'] = PersonFactory(address_book, u'Mail', keywords=kws)
            EMailAddressFactory(
                persons['p1'], u'icemac@example.net', set_as_default=True)
            persons['p2'] = PersonFactory(
                address_book, u'Mail Too', keywords=kws)
            EMailAddressFactory(
                persons['p2'], u'mail@example.com', set_as_default=True)
            persons['p3'] = PersonFactory(
                address_book, u'Double Mail', keywords=kws)
            # Dulicate email address
            EMailAddressFactory(
                persons['p3'], u'mail@example.com', set_as_default=True)
            persons['p4'] = FullPersonFactory(
                address_book, u'Other Mail', keywords=kws)
            # default email address is None
            EMailAddressFactory(
                persons['p4'], u'other@example.com', set_as_default=False)
            transaction.commit()
            yield connection, persons


# Tests

def test_mailto__MailTo__unique_mail_addresses__1(mailto_data):
    """It is a sorted list of unique email addresses."""
    from .mailto import MailTo
    from gocept.testing.mock import Property
    view = MailTo()
    view.request = mock.Mock()
    persons = ('icemac.addressbook.browser.search.result.handler.mailto.'
               'mailto.MailTo.persons')
    with mock.patch(persons, Property()) as persons,\
            zope.component.hooks.site(mailto_data['p0'].__parent__):
        persons.return_value = [mailto_data[x]
                                for x in 'p0 p1 p2 p3 p4'.split()]
        assert [u'icemac@example.net',
                u'mail@example.com'] == view.unique_mail_addresses


def test__mailto__MailTo___link__1(mailto_data, browser):
    """The view displays the e-mail adresses of the persons as a link."""
    browser.login('visitor')
    browser.keyword_search('mail-me', apply='E-Mail')
    assert [['<a href="mailto:?bcc=icemac@example.net,mail@example.com">'
             'Click here to open your e-mail client</a>'],
            ['<a href="javascript:history.go(-1)" '
             'class="no-print">Go back</a>']] == [
        browser.etree_to_list(x)
        for x in browser.etree.xpath('//div[@id="mailto"]//a')]



import zope.component
from icemac.addressbook.person import Person, Keywords
from icemac.addressbook.interfaces import IPerson, IPersonDefaults, IEntity
from icemac.addressbook.interfaces import IKeywordTitles, ITitle, IPhoneNumber
from icemac.addressbook.interfaces import IEntityOrder
import gocept.reference.verify


def test_person__Person__1():
    """It fulfills the `IPerson` interface."""
    assert gocept.reference.verify.verifyObject(IPerson, Person())


def test_person__Person__2():
    """It fulfills the `IPersonDefaults` interface."""
    assert gocept.reference.verify.verifyObject(IPersonDefaults, Person())


def test_person__Keywords__1():
    """It fulfills the `IKeywordTitles` interface."""
    assert gocept.reference.verify.verifyObject(IKeywordTitles, Keywords(None))


def test_person__PersonDefaultsEntity__getRawFields__1(address_book):
    """It returns the fields in default sort order if no order is defined."""
    entity = IEntity(IPersonDefaults)
    assert ([
        'default_postal_address',
        'default_phone_number',
        'default_email_address',
        'default_home_page_address',
    ] == [x[0] for x in entity.getRawFields()])


def test_person__PersonDefaultsEntity__getRawFields__2(address_book):
    """It returns the fields in the entity sort order."""
    entity = IEntity(IPersonDefaults)
    zope.component.getUtility(IEntityOrder).up(IEntity(IPhoneNumber))
    assert ([
        'default_phone_number',
        'default_postal_address',
        'default_email_address',
        'default_home_page_address',
    ] == [x[0] for x in entity.getRawFields()])


def test_person__PersonDefaultsEntity__getRawFields__3(address_book):
    """It returns the fields in default order if called whih `sorted=False`."""
    entity = IEntity(IPersonDefaults)
    zope.component.getUtility(IEntityOrder).up(IEntity(IPhoneNumber))
    assert ([
        'default_postal_address',
        'default_phone_number',
        'default_email_address',
        'default_home_page_address',
    ] == [x[0] for x in entity.getRawFields(sorted=False)])


def test_person__Person__get_name__1():
    """It returns last name and first name if both are set."""
    person = Person()
    person.first_name = u'Bernd'
    person.last_name = u'Tester'
    assert u'Bernd Tester' == person.get_name()


def test_person__Person__get_name__2():
    """It returns the last name if the first name is not set."""
    person = Person()
    person.last_name = u'Tester'
    assert u'Tester' == person.get_name()


def test_person__Person__get_name__3():
    """It returns the first name if the last name is not set."""
    person = Person()
    person.first_name = u'Berns'
    assert u'Berns' == person.get_name()


def test_person__Person__get_name__4():
    """It returns '' if neither first name nor last name is set."""
    assert u'' == Person().get_name()


def test_person__title__1(zcmlS):
    """The title of a person without a name is a fix string."""
    assert u'<no name>' == ITitle(Person())


def test_person__title__2(zcmlS):
    """`title()` returns the last name if the first name is empty."""
    person = Person()
    person.last_name = u'Tester'
    assert u'Tester' == ITitle(person)


def test_person__title__3(zcmlS):
    """`title()` returns the last name and first name comma separated."""
    person = Person()
    person.first_name = u'Hans'
    person.last_name = u'Tester'
    assert u'Tester, Hans' == ITitle(person)



from __future__ import absolute_import
from .interfaces import IBirthDate
from datetime import timedelta
import icalendar
import icemac.addressbook.browser.base
import icemac.addressbook.interfaces
import zope.interface
import icemac.addressbook.icalendar


ONE_DAY = timedelta(days=1)


class iCalendar(icemac.addressbook.browser.base.BaseView):
    """Render the person's birthdate as iCal file."""

    zope.interface.implements(IBirthDate)

    def __init__(self, context, request):
        # We need this contructor because this class is registered both as
        # browser:page and adapter!
        self.context = context
        self.request = request

    @property
    def icalendar_event(self):
        if self.context.birth_date is None:
            return
        summary = '%s (*%s)' % (
            icemac.addressbook.interfaces.ITitle(self.context),
            self.context.birth_date.year)
        return icalendar.Event(
            uid=self.url(self.context, 'iCalendar'),
            summary=summary,
            dtstart=icalendar.vDate(self.context.birth_date),
            dtend=icalendar.vDate(self.context.birth_date + ONE_DAY),
            rrule=icalendar.vRecur(freq='yearly')
        )

    def __call__(self):
        icemac.addressbook.icalendar.set_download_request_headers(
            self.request, 'birthday')
        cal = icemac.addressbook.icalendar.Calendar()
        cal.add_component(self.icalendar_event)
        return cal.to_ical()



from __future__ import absolute_import

from .base import SessionStorageStep
from icemac.addressbook.i18n import _
import icemac.addressbook.fieldsource
import zope.interface
import zope.schema


class IFieldChooser(zope.interface.Interface):
    """List of updateable fields."""

    field = zope.schema.Choice(
        title=_('Which field should be updated?'),
        source=icemac.addressbook.fieldsource.source)


class Field(SessionStorageStep):
    """Step where the user choosed the field he likes to update."""

    interface = IFieldChooser
    label = _(u'Choose field')



from __future__ import absolute_import

from .base import (
    SessionStorageStep, get_update_data_session, update_persons,
    clean_update_data_session, get_fieldname_in_session)
from icemac.addressbook.i18n import _
import icemac.addressbook.browser.personlist
import icemac.addressbook.browser.search.result.handler.base
import icemac.addressbook.browser.table
import icemac.addressbook.fieldsource
import icemac.addressbook.interfaces
import transaction
import z3c.form.field
import z3c.table.column
import zope.i18n


class TitleColumn(z3c.table.column.Column):
    """Column wich displays the title of the object."""

    def renderCell(self, person):
        return icemac.addressbook.interfaces.ITitle(person)


class ErrorColumn(z3c.table.column.Column):
    """Column displaying operation and validation errors."""

    def renderCell(self, person):
        session = get_update_data_session(self.request)
        return zope.i18n.translate(
            session.get('errors', {}).get(person.__name__, ''),
            context=self.request)


def get_chosen_entity_and_field(request):
    """Return entity and field objects for chosen field."""
    field_token = get_update_data_session(request)['field']
    return icemac.addressbook.fieldsource.untokenize(field_token)


class ReviewTable(icemac.addressbook.browser.table.Table,
                  icemac.addressbook.browser.search.result.handler.base.Base):
    """Table containing changed data for review."""

    sortOn = None

    def setUpColumns(self):
        entity, field = get_chosen_entity_and_field(self.request)
        return [
            z3c.table.column.addColumn(
                self, TitleColumn, 'name', weight=1, header=_('name')),
            icemac.addressbook.browser.personlist.createFieldColumn(
                self, entity, field, 2),
            z3c.table.column.addColumn(
                self, ErrorColumn, 'errors', weight=3, header=_('errors')),
        ]

    @property
    def values(self):
        return self.persons


class Result(SessionStorageStep,
             icemac.addressbook.browser.search.result.handler.base.Base):
    """Step where the user can check the result."""

    label = _(u'Check result')
    fields = z3c.form.field.Fields()
    handleApplyOnComplete = False

    @property
    def update_data(self):
        return get_update_data_session(self.request)

    def _completeable(self):
        """Tell whether this step can be completed."""
        update_data = self.update_data
        return 'field' in update_data and 'operation' in update_data

    def update(self):
        if self._completeable():
            # Otherwise the user has clicked on this step in navigation before
            # entring data.
            self._update_persons()
            # Make sure that changes are not yet persisted:
            transaction.doom()
        super(Result, self).update()

    def _update_persons(self):
        """Update the persons as the user selected it."""
        entity, field = get_chosen_entity_and_field(self.request)
        fieldname = self.getContent()['field']
        update_data = self.update_data
        errors = update_persons(
            self.persons, entity, field, update_data['operation'],
            update_data[get_fieldname_in_session(fieldname)])
        update_data['errors'] = errors

    def renderResultTable(self):
        table = ReviewTable(self.context, self.request)
        table.update()
        return table.render()

    @property
    def showCompleteButton(self):
        """Complete button condition."""
        if not self._completeable():
            return False
        if self.update_data.get('errors'):
            return False
        return super(Result, self).showCompleteButton

    def doComplete(self, action):
        """Handler for Complete button."""
        # Current transaction is doomed (see renderResultTable), so we have
        # to abort it to make it possible it will get commited at end of
        # request:
        transaction.abort()
        self._update_persons()
        clean_update_data_session(self.request)
        return super(Result, self).doComplete(action)



# -*- coding: utf-8 -*-
from icemac.addressbook.i18n import _
import ZODB.blob
import icemac.addressbook.entities
import icemac.addressbook.file.interfaces
import icemac.addressbook.interfaces
import persistent
import zope.container.contained
import zope.interface


class File(persistent.Persistent, zope.container.contained.Contained):
    """A file."""

    zope.interface.implements(icemac.addressbook.file.interfaces.IFile)
    zope.schema.fieldproperty.createFieldProperties(
        icemac.addressbook.file.interfaces.IFile, omit=['data', 'size'])

    def __init__(self, data='', *args, **kw):
        super(File, self).__init__(*args, **kw)
        # initialize blob with no data
        self._data = ZODB.blob.Blob()
        self.data = data

    @property
    def size(self):
        reader = self.open()
        reader.seek(0, 2)
        size = int(reader.tell())
        reader.close()
        return size

    @property
    def data(self):
        # This method can't be used to read the data, use openDetached instead.
        # This is necessary because of the stupidity of z3c.form which reads
        # the whole file when rendering a form.
        return ''

    @data.setter
    def data(self, data):
        fp = self.open('w')
        fp.write(data)
        fp.close()

    def replace(self, filename):
        """Replace with another file."""
        self._data.consumeFile(filename)

    def openDetached(self, mode='r'):
        return file(self._data.committed(), 'rb')

    def open(self, mode='r'):
        return self._data.open(mode)


file_entity = icemac.addressbook.entities.create_entity(
    _(u'file'), icemac.addressbook.file.interfaces.IFile, File)


@zope.component.adapter(icemac.addressbook.file.interfaces.IFile)
@zope.interface.implementer(icemac.addressbook.interfaces.ITitle)
def title(file):
    """Title of the file."""
    return file.name



# -*- coding: utf-8 -*-
import z3c.form.browser.orderedselect
import z3c.form.widget


def sort_key(value):
    """Sort key for widget."""
    return value['content'].lower()


class SortedSelectWidget(z3c.form.browser.orderedselect.OrderedSelectWidget):
    """"OrderedSelectWidget which sorts the selected values itself."""

    def update(self):
        super(SortedSelectWidget, self).update()
        self.selectedItems = sorted(self.selectedItems, key=sort_key)
        self.notselectedItems = sorted(self.notselectedItems, key=sort_key)


def SortedSelectFieldWidget(field, request):
    """`IFieldWidget` factory for SortedSelectWidget."""
    return z3c.form.widget.FieldWidget(field, SortedSelectWidget(request))



import zope.interface


class IBirthDate(zope.interface.Interface):
    """Person's birthdate data."""

    icalendar_event = zope.interface.Attribute(
        'Birthdate as icalendar.Event or `None`')



# -*- coding: latin-1 -*-
from icemac.addressbook.i18n import _
import gocept.reference.interfaces
import grokcore.component
import icemac.addressbook.browser.interfaces
import icemac.addressbook.browser.resource
import icemac.addressbook.interfaces
import icemac.addressbook.utils
import transaction
import urllib
import z3c.flashmessage.interfaces
import z3c.form.button
import z3c.form.field
import z3c.form.group
import z3c.formui.form
import zope.component
import zope.interface
import zope.security
import zope.security.interfaces
import zope.security.proxy
import zope.session.interfaces
import zope.component.hooks
import zope.traversing.api
import zope.traversing.browser
import zope.traversing.publicationtraverse


def create(form, class_, data):
    """Create an object from a class and assign values."""
    obj = icemac.addressbook.utils.create_obj(class_)
    z3c.formui.form.applyChanges(form, obj, data)
    return obj


def all_(*constraints):
    """All given button constraints must evaluate to true."""
    return lambda form: all(constraint(form) for constraint in constraints)


class FlashView(object):
    """Base class to send flash messages."""

    def send_flash(self, message):
        zope.component.getUtility(
            z3c.flashmessage.interfaces.IMessageSource).send(message)


class BaseView(FlashView):
    """Base for view classes."""

    __PACKAGE_ID__ = icemac.addressbook.interfaces.PACKAGE_ID

    def url(self, obj, view_name=None, **kw):
        url_parts = [zope.traversing.browser.absoluteURL(obj, self.request)]
        if view_name:
            url_parts.append('/')
            if not view_name.startswith('++'):
                url_parts.append('@@')
            url_parts.append(view_name)
        if kw:
            url_parts.extend(['?', urllib.urlencode(kw, doseq=True)])
        return ''.join(url_parts)

    @property
    def session(self):
        return get_session(self.request, name=self.__PACKAGE_ID__)


class BaseForm(BaseView):
    """Base for all forms."""

    interface = None  # interface for form

    # privat
    _status = ''
    _fields = None

    @property
    def fields(self):
        if self._fields is not None:
            return self._fields
        if self.interface is None:
            field_values = ()
        else:
            fields = icemac.addressbook.interfaces.IEntity(self.interface)
            field_values = fields.getFieldValues()
        self._fields = z3c.form.field.Fields(*field_values)
        return self._fields

    @fields.setter
    def fields(self, fields):
        if self._fields is not None:
            raise ValueError('fields is already set')
        self._fields = fields

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, message):
        if message != z3c.form.form.Form.formErrorsMessage:
            # Non-error messages must be handled by z3c.flashmessage, too, as
            # they must be displayed on the next page after the redirect. But
            # we need them to determine whether a redirect is necessary, too.
            self.send_flash(message)
        self._status = message


class BaseAddForm(BaseForm, z3c.formui.form.AddForm):
    """Simple base add form."""

    class_ = None  # create object from this class
    next_url = None  # target after creation, one of ('object', 'parent')
    next_view = None  # target view after creation, None --> default view
    _next_url_map = {
        'object': lambda self: self.context[self._name],
        'parent': lambda self: self.context
    }

    def createAndAdd(self, data):
        # overwriting method as otherwise object created event would
        # get send twice
        self.obj = self.create(data)
        self.add(self.obj)
        return self.obj

    def create(self, data):
        return create(self, self.class_, data)

    def add(self, obj):
        try:
            self._name = icemac.addressbook.utils.add(self.context, obj)
        except zope.interface.Invalid, e:
            transaction.doom()
            raise z3c.form.interfaces.ActionExecutionError(e)

    @z3c.form.button.buttonAndHandler(_('Add'), name='add')
    def handleAdd(self, action):
        super(BaseAddForm, self).handleAdd(self, action)
        if self._finishedAdd:
            self.status = _(
                '"${title}" added.',
                mapping=dict(title=icemac.addressbook.interfaces.ITitle(
                    self.obj)))

    @z3c.form.button.buttonAndHandler(_('Cancel'), name='cancel')
    def handleCancel(self, action):
        self.next_url = 'parent'
        self._finishedAdd = True
        self.status = _('Addition canceled.')

    def nextURL(self):
        return self.url(self._next_url_map[self.next_url](self),
                        self.next_view)


def update_with_redirect(class_, self):
    """Call update of super class and redirect when necessary."""
    # Caution: we need the class_ parameter as we get infinite recursion if
    # using self.__class__ in the super call.
    super(class_, self).update()
    if self.request.response.getStatus() in (302, 303, 304):
        # already redirecting
        return
    if self.status in (self.successMessage, self.noChangesMessage):
        self.redirect_to_next_url()


class _AbstractEditForm(BaseForm, z3c.formui.form.EditForm):
    """Abstract base class for edit forms.

    CAUTION: Not to be used as direct base class of forms, use one of its
             child classes!
    """

    next_url = None  # target object after edit, one of ('object', 'parent')
    next_view = None  # target view after edit (None for default view)
    id = 'edit-form'
    _next_url_map = {
        'object': lambda self: self.context,
        'parent': lambda self: zope.traversing.api.getParent(self.context),
        'site': lambda self: zope.component.hooks.getSite()
    }

    def update(self):
        update_with_redirect(_AbstractEditForm, self)

    def redirect_to_next_url(self, next_url=None, next_view=None):
        target = self._next_url_map[
            next_url if next_url else self.next_url](self)

        if next_view is None:
            next_view = self.next_view
        self.request.response.redirect(self.url(target, next_view))

    def applyChanges(self, data):
        try:
            return super(_AbstractEditForm, self).applyChanges(data)
        except zope.interface.Invalid, e:
            transaction.doom()
            raise z3c.form.interfaces.ActionExecutionError(e)


class BaseEditForm(_AbstractEditForm):
    """Base edit form.

    It has a cancel button registered on it.
    """


class EditActions(z3c.form.button.ButtonActions,
                  grokcore.component.MultiAdapter):
    """Custom edit actions to add a cancel button on each edit form."""

    grokcore.component.adapts(BaseEditForm,
                              zope.interface.Interface,
                              zope.interface.Interface)

    def update(self):
        self.form.buttons = z3c.form.button.Buttons(
            self.form.buttons,
            z3c.form.button.Button('cancel', _(u'Cancel')))
        super(EditActions, self).update()


class EditActionHandler(z3c.form.button.ButtonActionHandler,
                        grokcore.component.MultiAdapter):
    """Edit action handler which is able to handle the cancel action."""

    grokcore.component.adapts(BaseEditForm,
                              zope.interface.Interface,
                              zope.interface.Interface,
                              z3c.form.button.ButtonAction)

    def __call__(self):
        if self.action.name == 'form.buttons.cancel':
            self.form.status = self.form.noChangesMessage
            return
        super(EditActionHandler, self).__call__()


class GroupEditForm(z3c.form.group.GroupForm, BaseEditForm):
    """EditForm (with cancel) as group form."""

    # This is needed here as GroupForm does not do a super-call.
    def update(self):
        update_with_redirect(GroupEditForm, self)


class _BaseConfirmForm(_AbstractEditForm):
    """Display a confirmation dialog before the action."""

    label = NotImplemented
    interface = None
    field_names = ()  # tuple of field names for display; empty for all
    cancel_status_message = NotImplemented

    requiredInfo = None  # do never display requiredInfo
    mode = z3c.form.interfaces.DISPLAY_MODE
    next_url_after_cancel = 'object'

    def update(self):
        icemac.addressbook.browser.resource.no_max_content_css.need()
        super(_BaseConfirmForm, self).update()

    @property
    def fields(self):
        fields = super(_BaseConfirmForm, self).fields
        if self.field_names:
            fields = fields.select(*self.field_names)
        return fields

    @z3c.form.button.buttonAndHandler(_(u'No, cancel'), name='cancel')
    def handleCancel(self, action):
        self.redirect_to_next_url(self.next_url_after_cancel)
        self.status = self.cancel_status_message

    @z3c.form.button.buttonAndHandler(_(u'Yes'), name='action')
    def handleAction(self, action):
        self._handle_action()


class BaseDeleteForm(_BaseConfirmForm):
    """Display a deletion confirmation dialog."""

    label = _(u'Do you really want to delete this entry?')
    next_view_after_delete = None  # None --> default view
    next_url_after_delete = 'parent'
    cancel_status_message = _('Deletion canceled.')

    def _handle_action(self):
        self.redirect_to_next_url(self.next_url_after_delete,
                                  self.next_view_after_delete)
        self.status = _('"${title}" deleted.',
                        mapping={'title': self.status_title})
        self._do_delete()

    def _do_delete(self):
        icemac.addressbook.utils.delete(self.context)

    @property
    def status_title(self):
        return icemac.addressbook.interfaces.ITitle(self.getContent())


def delete_persons(address_book, ids):
    """Delete persons specified by their ID, but not users."""
    deleted = 0
    for name in list(ids):  # this list() call is needed as we might delete
                            # from the source of the ids
        ref_target = gocept.reference.interfaces.IReferenceTarget(
            zope.security.proxy.getObject(address_book[name]))
        if ref_target.is_referenced(recursive=False):
            # Persons which are referenced can't be deleted using this
            # function. We check this here to avoid getting an error.
            continue
        del address_book[name]
        deleted += 1
    return deleted


class PrefixGroup(z3c.form.group.Group):
    """Group which sets a prefix."""

    prefix = None  # to be set in subclass
    interface = None  # to be set in subclass

    @property
    def fields(self):
        fields = icemac.addressbook.interfaces.IEntity(self.interface)
        return z3c.form.field.Fields(
            *fields.getFieldValues(), **dict(prefix=self.prefix))


class BaseCloneForm(_BaseConfirmForm):
    """Display a cloning confirmation dialog."""

    label = _(u'Do you really want to clone this entry?')
    cancel_status_message = _('Cloning canceled.')

    def _handle_action(self):
        # clone object
        unsecure_context = zope.security.proxy.getObject(self.context)
        copier = zope.copypastemove.interfaces.IObjectCopier(unsecure_context)
        parent = self.context.__parent__
        new_name = copier.copyTo(parent)
        # redirect to clone
        self.request.response.redirect(self.url(parent[new_name]))
        zope.component.getUtility(
            z3c.flashmessage.interfaces.IMessageSource).send(
            _('"${object}" cloned.', mapping=dict(
                object=icemac.addressbook.interfaces.ITitle(self.context))))


def can_access_uri_part(context, request, uri_part):
    """Tell if current user can to access a URI part relative to context."""
    assert uri_part is not None
    traverser = zope.traversing.publicationtraverse.PublicationTraverser()
    try:
        view = traverser.traverseRelativeURL(request, context, uri_part)
    except (zope.security.interfaces.Unauthorized,
            zope.security.interfaces.Forbidden,
            LookupError):
        return False
    # We're assuming that view pages are callable this is a pretty sound
    # assumption:
    return zope.security.canAccess(view, '__call__')


def can_access(uri_part):
    """Button condition function to if access is allowed.

    Returns a function which tests whether the user can access the URL
    ``context/@@absolute_url/<uri_part>``.
    """
    def can_access_form(form):
        return can_access_uri_part(form.context, form.request, uri_part)
    return can_access_form


def get_session(request, name=icemac.addressbook.interfaces.PACKAGE_ID):
    """Get the browser session of the current user."""
    return zope.session.interfaces.ISession(request)[name]



from ..base import Base
from icemac.addressbook.browser.person.interfaces import IBirthDate
from icemac.addressbook.i18n import _
import icemac.addressbook.icalendar
import icemac.addressbook.interfaces
import zope.component
import zope.i18n


class iCalendar(Base):
    """Export birthdays of selected persons as iCalendar file."""

    def __call__(self):
        icemac.addressbook.icalendar.set_download_request_headers(
            self.request, 'birthdays')
        exporter = Exporter(self.persons, self.request)
        return exporter()


class Exporter(object):

    def __init__(self, persons, request):
        self.persons = persons
        self.request = request
        ab_title = icemac.addressbook.interfaces.IAddressBook(None).title
        message_id = _('birthdays of ${address_book}',
                       mapping={'address_book': ab_title})
        cal_name = zope.i18n.translate(message_id, context=self.request)
        # x-wr-calname is the calendar name used by Apple's iCal
        self.calendar = icemac.addressbook.icalendar.Calendar(
            **{'x-wr-calname': cal_name})

    def update(self):
        for person in self.persons:
            birthdate_data = zope.component.getMultiAdapter(
                (person, self.request), IBirthDate)
            event = birthdate_data.icalendar_event
            if event:
                self.calendar.add_component(event)

    def __call__(self):
        self.update()
        return self.calendar.to_ical()



# -*- coding: utf-8 -*-
import grokcore.component as grok
import icemac.addressbook.browser.interfaces
import icemac.addressbook.browser.menus.menu
import z3c.menu.ready2go.checker
import z3c.menu.ready2go.item
import zope.interface
import zope.preference.interfaces


class PreferencesMenuItem(z3c.menu.ready2go.item.SiteMenuItem):
    """Menu item for the preferences tab in the site menu."""


class PreferencesMenuItemSelectedChecker(
        z3c.menu.ready2go.checker.TrueSelectedChecker,
        grok.MultiAdapter):
    """Selected checker for the Preferences menu item in the site menu."""

    grok.adapts(zope.interface.Interface,
                icemac.addressbook.browser.interfaces.IAddressBookLayer,
                zope.interface.Interface,
                icemac.addressbook.browser.menus.menu.MainMenu,
                PreferencesMenuItem)

    @property
    def selected(self):
        return zope.preference.interfaces.IPreferenceGroup.providedBy(
            self.context)



from zope.testbrowser.browser import HTTPError


def test_root__FrontPage__1(empty_zodb, browser):
    """`FrontPage` shows a message if there is no address book created, yet."""
    browser.login('mgr')
    browser.open(browser.ROOT_URL)
    print browser.contents
    assert ('There are no address books created yet, click on "Add address '
            'book" to create one' in browser.contents)


def test_root__FrontPage__2(empty_zodb, browser):
    """`FrontPage` cannot be accessed by anonymous."""
    import pytest
    with pytest.raises(HTTPError) as err:
        browser.open(browser.ROOT_URL)
    assert 'HTTP Error 401: Unauthorized' == str(err.value)


def test_root__FrontPage__3(address_book, browser):
    """`FrontPage` lists only address books.

    When another object is in the root folder it is not listed.
    """
    from zope.container.btree import BTreeContainer
    address_book.__parent__['btree'] = BTreeContainer()
    browser.login('mgr')
    browser.open(browser.ROOT_URL)
    assert ['test addressbook'] == browser.etree.xpath('//ul/li/a[1]/text()')


def test_root__FrontPage__4(
        address_book, AddressBookFactory, PersonFactory, browser):
    """`FrontPage` shows number of persons in the address book."""
    ab2 = AddressBookFactory('ab2', u'ab with content')
    ab3 = AddressBookFactory('ab3', u'ab with content, too')
    PersonFactory(ab2, u'Tester')
    PersonFactory(ab3, u'Tester2')
    PersonFactory(ab3, u'Tester3')
    browser.login('mgr')
    browser.open(browser.ROOT_URL)
    assert (['(0 items)', '(1 item)', '(2 items)'] ==
            [x.strip()
             for x in browser.etree.xpath('//ul[@class="bullet"]/li/text()')
             if x.strip()])



# -*- coding: utf-8 -*-
import xlrd
from icemac.addressbook.interfaces import IPerson


def create_person(
        field_title, keyword_title, address_book, FieldFactory,
        FullPersonFactory, PostalAddressFactory):
    """Create a person in the address book as needed by the tests."""
    field_name = FieldFactory(
        address_book, IPerson, u'Bool', field_title).__name__
    person = FullPersonFactory(
        address_book, u'Liebig', postal__city=u'Testhausen',
        keywords=[keyword_title], **{field_name: True})
    # Let's create an additional postal address which does not show up in
    # the defaults export but in complete export:
    PostalAddressFactory(person, zip=u'00001')


def test_xls__DefaultsExport__1(search_data, browser):
    """`DefaultsExport` can export the values of the default data as XLS."""
    browser.login('visitor')
    browser.keyword_search('church', 'XLS export main')
    assert 'application/vnd.ms-excel' == browser.headers['Content-Type']
    assert ('attachment; filename=addressbook_export.xls' ==
            browser.headers['Content-Disposition'])
    xls_workbook = xlrd.open_workbook(file_contents=browser.contents)
    assert [u'Address book - Export'] == xls_workbook.sheet_names()
    work_sheet_0 = xls_workbook.sheet_by_index(0)
    assert (5, 13) == (work_sheet_0.nrows, work_sheet_0.ncols)
    assert [
        [u'person',
         '',
         '',
         '',
         '',
         u'postal address',
         '',
         '',
         '',
         '',
         u'phone number',
         u'e-mail address',
         u'home page address'],
        [u'first name',
         u'last name',
         u'birth date',
         u'keywords',
         u'notes',
         u'address prefix',
         u'street',
         u'city',
         u'zip',
         u'country',
         u'number',
         u'e-mail address',
         u'URL'],
        ['',
         u'Koch',
         19017.0,
         u'family, church',
         u'father-in-law',
         '',
         '',
         '',
         '',
         u'Germany',
         '',
         '',
         ''],
        ['',
         u'Liebig',
         '',
         u'church',
         u'family',
         '',
         '',
         '',
         '',
         u'Germany',
         '',
         '',
         ''],
        ['',
         u'Velleuer',
         '',
         u'family, church',
         '',
         '',
         '',
         '',
         '',
         u'Germany',
         '',
         '',
         '']] == [work_sheet_0.row_values(rx)
                  for rx in range(work_sheet_0.nrows)]


def test_xls__DefaultsExport__2(search_data, browser):
    """`DefaultsExport` only exports the selected persons."""
    browser.login('visitor')
    browser.keyword_search('church')
    browser.getControl('Apply on selected persons').displayValue = [
        'XLS export main']
    browser.getControl(
        name='persons:list').getControl(value="Person-2").selected = False
    browser.getControl(name='form.buttons.apply').click()
    xls_workbook = xlrd.open_workbook(file_contents=browser.contents)
    work_sheet_0 = xls_workbook.sheet_by_index(0)
    assert (4, 13) == (work_sheet_0.nrows, work_sheet_0.ncols)
    assert [
        [u'person',
         '',
         '',
         '',
         '',
         u'postal address',
         '',
         '',
         '',
         '',
         u'phone number',
         u'e-mail address',
         u'home page address'],
        [u'first name',
         u'last name',
         u'birth date',
         u'keywords',
         u'notes',
         u'address prefix',
         u'street',
         u'city',
         u'zip',
         u'country',
         u'number',
         u'e-mail address',
         u'URL'],
        ['',
         u'Liebig',
         '',
         u'church',
         u'family',
         '',
         '',
         '',
         '',
         u'Germany',
         '',
         '',
         ''],
        ['',
         u'Velleuer',
         '',
         u'family, church',
         '',
         '',
         '',
         '',
         '',
         u'Germany',
         '',
         '',
         '']] == [work_sheet_0.row_values(rx)
                  for rx in range(work_sheet_0.nrows)]


def test_xls__DefaultsExport__3(
        address_book, FieldFactory, FullPersonFactory, PostalAddressFactory,
        browser):
    """`DefaultsExport` exports user defined fields like other fields."""
    create_person(u'photo permission?', u'church', address_book, FieldFactory,
                  FullPersonFactory, PostalAddressFactory)
    browser.login('visitor')
    browser.open(browser.SEARCH_URL)
    # We choose the keyword search as it has export abilities:
    browser.getLink('Keyword search').click()
    browser.getControl('keywords').displayValue = ['church']
    browser.getControl('Search').click()
    # The export produces an XLS file:
    browser.getControl('Apply on selected persons').displayValue = [
        'XLS export main']
    browser.getControl(name='form.buttons.apply').click()
    assert 'application/vnd.ms-excel' == browser.headers['Content-Type']
    xls_workbook = xlrd.open_workbook(file_contents=browser.contents)
    work_sheet_0 = xls_workbook.sheet_by_index(0)
    assert (3, 14) == (work_sheet_0.nrows, work_sheet_0.ncols)
    assert [
        [u'person',
         '',
         '',
         '',
         '',
         '',
         u'postal address',
         '',
         '',
         '',
         '',
         u'phone number',
         u'e-mail address',
         u'home page address'],
        [u'first name',
         u'last name',
         u'birth date',
         u'keywords',
         u'notes',
         u'photo permission?',
         u'address prefix',
         u'street',
         u'city',
         u'zip',
         u'country',
         u'number',
         u'e-mail address',
         u'URL'],
        ['',
         u'Liebig',
         '',
         u'church',
         '',
         True,
         '',
         '',
         u'Testhausen',
         '',
         u'Germany',
         '',
         '',
         '']] == [work_sheet_0.row_values(rx)
                  for rx in range(work_sheet_0.nrows)]


def test_xls__DefaultsExport__4(
        translated_address_book, FieldFactory, FullPersonFactory,
        PostalAddressFactory, browser):
    """`DefaultsExport` translates field names into user's language."""
    create_person(u'Fotoerlaubnis?', u'Kirche', translated_address_book,
                  FieldFactory, FullPersonFactory, PostalAddressFactory)
    # As visitors are allowed to search and export, so we log in as a
    # visitor, to enable translation, we also send an accept-language
    # header:
    browser.login('visitor')
    browser.lang('de-DE')
    browser.open(browser.SEARCH_BY_KEYWORD_URL)
    browser.getControl('Schlagwörter').displayValue = ['Kirche']
    browser.getControl('Suchen').click()
    browser.getControl('Auf ausgewählte Personen anwenden').displayValue = [
        'wichtigste']
    browser.getControl(name='form.buttons.apply').click()
    assert 'application/vnd.ms-excel' == browser.headers['Content-Type']
    xls_workbook = xlrd.open_workbook(file_contents=browser.contents)
    work_sheet_0 = xls_workbook.sheet_by_index(0)
    assert (3, 14) == (work_sheet_0.nrows, work_sheet_0.ncols)
    assert [
        [u'Person',
         '',
         '',
         '',
         '',
         '',
         u'Anschrift',
         '',
         '',
         '',
         '',
         u'Telefonnummer',
         u'E-Mail-Adresse',
         u'Homepage-Adresse'],
        [u'Vorname',
         u'Familienname',
         u'Geburtsdatum',
         u'Schlagwörter',
         u'Anmerkungen',
         u'Fotoerlaubnis?',
         u'Adresszusatz',
         u'Straße',
         u'Ort',
         u'PLZ',
         u'Land',
         u'Nummer',
         u'E-Mail-Adresse',
         u'URL'],
        ['',
         u'Liebig',
         '',
         u'Kirche',
         '',
         1,
         '',
         '',
         u'Testhausen',
         '',
         u'Deutschland',
         '',
         '',
         '']] == [work_sheet_0.row_values(rx)
                  for rx in range(work_sheet_0.nrows)]


def test_xls__CompleteExport__1(
        translated_address_book, FieldFactory, FullPersonFactory,
        PostalAddressFactory, browser):
    """`CompleteExport` translates field names into user's language."""
    create_person(u'Fotoerlaubnis?', u'Kirche', translated_address_book,
                  FieldFactory, FullPersonFactory, PostalAddressFactory)
    # The complete export is not really different from the main one:
    browser.login('visitor')
    browser.lang('de-DE')
    browser.open(browser.SEARCH_BY_KEYWORD_URL)
    browser.getControl('Schlagwörter').displayValue = ['Kirche']
    browser.getControl('Suchen').click()
    browser.getControl('Auf ausgewählte Personen anwenden').displayValue = [
        'vollständig']
    browser.getControl(name='form.buttons.apply').click()
    assert 'application/vnd.ms-excel' == browser.headers['Content-Type']
    xls_workbook = xlrd.open_workbook(file_contents=browser.contents)
    work_sheet_0 = xls_workbook.sheet_by_index(0)
    assert (3, 19) == (work_sheet_0.nrows, work_sheet_0.ncols)
    assert [
        [u'Person',
         '',
         '',
         '',
         '',
         '',
         u'wichtigste Anschrift',
         '',
         '',
         '',
         '',
         u'weitere Anschrift',
         '',
         '',
         '',
         '',
         u'wichtigste Telefonnummer',
         u'wichtigste E-Mail-Adresse',
         u'wichtigste Homepage-Adresse'],
        [u'Vorname',
         u'Familienname',
         u'Geburtsdatum',
         u'Schlagwörter',
         u'Anmerkungen',
         u'Fotoerlaubnis?',
         u'Adresszusatz',
         u'Straße',
         u'Ort',
         u'PLZ',
         u'Land',
         u'Adresszusatz',
         u'Straße',
         u'Ort',
         u'PLZ',
         u'Land',
         u'Nummer',
         u'E-Mail-Adresse',
         u'URL'],
        ['',
         u'Liebig',
         '',
         u'Kirche',
         '',
         1,
         '',
         '',
         u'Testhausen',
         '',
         u'Deutschland',
         '',
         '',
         '',
         u'00001',
         u'Deutschland',
         '',
         '',
         '']] == [work_sheet_0.row_values(rx)
                  for rx in range(work_sheet_0.nrows)]


def test_xls__CompleteExport__2(search_data, browser):
    """`CompleteExport` does not fail if no person is selected."""
    browser.login('visitor')
    browser.keyword_search('friends')
    browser.getControl('Apply on selected persons').displayValue = [
        'XLS export complete']
    browser.getControl(
        name='persons:list').getControl(value="Person").selected = False
    browser.getControl(name='form.buttons.apply').click()
    # When the user chooses no person for export a nearly empty sheet gets
    # exported.
    xls_workbook = xlrd.open_workbook(file_contents=browser.contents)
    work_sheet_0 = xls_workbook.sheet_by_index(0)
    assert (2, 5) == (work_sheet_0.nrows, work_sheet_0.ncols)
    assert [
        [u'person', '', '', '', ''],
        [u'first name',
         u'last name',
         u'birth date',
         u'keywords',
         u'notes']] == [work_sheet_0.row_values(rx)
                        for rx in range(work_sheet_0.nrows)]



# -*- coding: utf-8 -*-
from StringIO import StringIO
from icemac.addressbook.install import not_matched_prerequisites, Configurator
import contextlib
import os
import pytest
import sys


# Fixtures


@pytest.fixture('function')
def basedir(tmpdir):
    """Create a base directory for the tests and chdir to it."""
    cwd = os.getcwd()
    tmpdir.chdir()
    yield tmpdir
    os.chdir(cwd)


@pytest.fixture(scope='function')
def install_default_ini(basedir):
    """Create an `install.default.ini` for `Configurator`."""
    install_default_ini = basedir.join('install.default.ini')
    install_default_ini.write("""\
[install]
eggs_dir = py-eggs

[admin]
login = me
password = secret

[server]
host = my.computer.local
port = 13090
username =

[log]
handler = FileHandler
max_size = 1000
when = midnight
interval = 1
backups = 5

[packages]

[migration]
do_migration = no
stop_server = no
start_server = no
""")


@pytest.fixture(scope='function')
def config(install_default_ini):
    """Return a `Configurator` instance running on `install.default.ini`."""
    config = Configurator(stdin=StringIO())
    config.load()
    return config


@contextlib.contextmanager
def user_input(input, config):
    r"""Write `input` on `stdin`.

    If `input` is a list, join it using `\n`.
    """
    # Remove possibly existing previous input:
    config.stdin.seek(0)
    config.stdin.truncate()
    if not isinstance(input, basestring):
        input = '\n'.join(input)
    config.stdin.write(input)
    config.stdin.seek(0)
    yield


def test_install__not_matched_prerequisites__1(basedir):
    """It returns an error text if `buildout.cfg` already exists in `cwd`."""
    buildout_cfg = basedir.join('buildout.cfg')
    buildout_cfg.write('[buildout]')
    assert ('ERROR: buildout.cfg already exists.\n'
            '       Please (re-)move the existing one and restart install.' ==
            not_matched_prerequisites())


def test_install__not_matched_prerequisites__2(basedir):
    """It returns `False`:

    * if no `buildout.cfg` exists in `cwd` and
    * if the right python version is used.

    We expect that the version of the python which runs the tests matches the
    requirement.
    """
    assert False is not_matched_prerequisites()


def test_install__not_matched_prerequisites__3(monkeypatch, basedir):
    """It returns an error text for a too old python version.

    `icemac.addressbook` currently only runs with some Python versions. If
    another version is used, an error message is returned.
    """
    monkeypatch.setattr("sys.version_info", (2, 5, 6, 'final', 0))
    assert ('ERROR: icemac.addressbook currently supports only Python 2.7.'
            '\n       But you try to install it using Python 2.5.6.' ==
            not_matched_prerequisites())


def test_install__not_matched_prerequisites__4(monkeypatch, basedir):
    """It returns an error text for a too new python version."""
    monkeypatch.setattr("sys.version_info", (3, 0, 0, 'final', 0))
    assert ('ERROR: icemac.addressbook currently supports only Python 2.7.'
            '\n       But you try to install it using Python 3.0.0.' ==
            not_matched_prerequisites())


def test_install__Configurator__1(config):
    """It reads the default values from `install.default.ini`.

    An instance of this class reads the default configuration file
    (`install.default.ini`). Whose values are used when the user does not enter
    a value.
    """
    assert 'py-eggs' == config.get('install', 'eggs_dir')


def test_install__Configurator__2(config):
    """It stores the path to the old instance. By default it is empty."""
    assert '' == config.get('migration', 'old_instance')


def test_install__Configurator__3():
    """It uses `sys.stdin` if it is not explicitly set."""
    config = Configurator()
    assert sys.stdin == config.stdin


def test_install__Configurator__ask_user__1(config, capsys):
    """It asks for input from the user.

    It displays the default value as set in the configuration file. The entered
    value is stored in the configuration and is also returned. When the user
    does not enter a value (only hits `enter`), the default value is used
    instead.
    """
    with user_input('\n', config):
        config.ask_user('Server port', 'server', 'port')
    assert (u' Server port: [13090] \n', u'') == capsys.readouterr()
    assert '13090' == config.get('server', 'port')


def test_install__Configurator__ask_user__3(config, capsys):
    """It prefers the user entered value over the one in default config."""
    with user_input('4711', config):
        config.ask_user('Server port', 'server', 'port')
    with user_input('0815', config):
        config.ask_user('Server port', 'server', 'port')
    assert ((u' Server port: [13090] \n Server port: [4711] \n', u'') ==
            capsys.readouterr())
    assert '0815' == config.get('server', 'port')


def test_install__Configurator__ask_user__4(config, capsys):
    """It allows force entering a value of a list.

    It retries to get an answer until the user entered a value is contained in
    the list of allowed values.
    """
    with user_input(['maybe', '???', 'yes'], config):
        config.ask_user(
            'Really', 'migration', 'do_migration', values=('yes', 'no'))
    assert ([u' Really: [no] ',
             u"ERROR: 'maybe' is not in ('yes', 'no').",
             u'Please choose a value out of the list.',
             u' Really: [no] ',
             u"ERROR: '???' is not in ('yes', 'no').",
             u'Please choose a value out of the list.',
             u' Really: [no] '] == capsys.readouterr()[0].splitlines())
    assert 'yes' == config.get('migration', 'do_migration')


def test_install__Configurator__load__1(config):
    """It (Re-) loads the configuration from the configuration file(s)."""
    with user_input('4711', config):
        config.ask_user('Server port', 'server', 'port')
    assert '4711' == config.get('server', 'port')
    config.load()
    assert '13090' == config.get('server', 'port')


def test_install__Configurator__load__2(config):
    """It raises an error if `user_config` points to a not existing file."""
    config.user_config = 'i-do-not-exist.ini'
    with pytest.raises(IOError) as err:
        config.load()
    assert "'i-do-not-exist.ini' does not exist." == str(err.value)


def test_install__Configurator__load__3(config, basedir):
    """It assures that user configuration takes precedence over default."""
    assert 'py-eggs' == config.get('install', 'eggs_dir')
    user_ini = basedir.mkdir('prev_version').join('install.user.ini')
    user_ini.write("[install]\neggs_dir = my-eggs")
    config.user_config = str(user_ini)
    config.load()
    assert 'my-eggs' == config.get('install', 'eggs_dir')
    assert (str(basedir.join('prev_version')) ==
            config.get('migration', 'old_instance'))


def test_install__Configurator__load__4(config, basedir):
    """Resetting the user_config to `None` clears `old_instance` path."""
    user_ini = basedir.mkdir('prev_version').join('install.user.ini')
    user_ini.write('')

    config.user_config = str(user_ini)
    config.load()
    assert '' != config.get('migration', 'old_instance')

    config.user_config = None
    config.load()
    assert '' == config.get('migration', 'old_instance')


def test_install__Configurator__print_intro__1(config, capsys):
    """It prints an introduction text to stdout."""
    config.print_intro()
    assert ([
        u'Welcome to icemac.addressbook installation',
        u'',
        u'Hint: to use the default value (the one in [brackets]), '
        u' enter no value.',
        u''] == capsys.readouterr()[0].splitlines())


def test_install__Configurator__get_global_options__1(config, capsys):
    """It ask the user about global options and stores them."""
    with user_input('/Users/mac', config):
        config.get_global_options()
    assert (' Directory to store python eggs: [py-eggs] \n' ==
            capsys.readouterr()[0])
    assert '/Users/mac' == config.eggs_dir
    assert '/Users/mac' == config.get('install', 'eggs_dir')


def test_install__Configurator__get_server_options__1(config, capsys):
    """It asks the user about the server and store the answers."""
    with user_input(['admin', 'geheim', 'localhost', '8080', 'mac'], config):
        config.get_server_options()
    assert ([
        u' Log-in name for the administrator: [me] ',
        u' Password for the administrator: [secret] ',
        u' Hostname: [my.computer.local] ',
        u' Port number: [13090] ',
        u' Username whether process should run as different user otherwise '
        u'emtpy: [] ',
    ] == capsys.readouterr()[0].splitlines())
    assert 'admin' == config.admin_login
    assert 'admin' == config.get('admin', 'login')
    assert 'geheim' == config.admin_passwd
    assert 'geheim' == config.get('admin', 'password')
    assert 'localhost' == config.host
    assert 'localhost' == config.get('server', 'host')
    assert '8080' == config.port
    assert '8080' == config.get('server', 'port')
    assert 'mac' == config.username
    assert 'mac' == config.get('server', 'username')


def test_install__Configurator__get_log_options__1(config, capsys):
    """It asks the user about logging and store his answers.

    When the user enters a wrong handler name he is asked to enter it again.
    Choosing `TimedRotatingFileHandler` three additional questions are asked.
    """
    with user_input(['does not matter', 'TimedRotatingFileHandler', 'D', '7',
                     '8'], config):
        config.get_log_options()
    assert ([
        u' Please choose Log-Handler:',
        u'    Details see http://docs.python.org/library/logging.html#'
        u'handler-objects',
        u' Log-Handler, choose between FileHandler, RotatingFileHandler, '
        u'TimedRotatingFileHandler: [FileHandler] ',
        u"ERROR: 'does not matter' is not in ('FileHandler', "
        u"'RotatingFileHandler', 'TimedRotatingFileHandler').",
        u'Please choose a value out of the list.',
        u' Log-Handler, choose between FileHandler, RotatingFileHandler, '
        u'TimedRotatingFileHandler: [FileHandler] ',
        u' Type of rotation interval, choose between S, M, H, D, W, '
        u'midnight: [midnight] ',
        u' Rotation interval size: [1] ',
        u' Number of log file backups: [5] ',
    ] == capsys.readouterr()[0].splitlines())
    assert 'TimedRotatingFileHandler' == config.log_handler
    assert 'TimedRotatingFileHandler' == config.get('log', 'handler')
    assert 'D' == config.log_when
    assert 'D' == config.get('log', 'when')
    assert '7' == config.log_interval
    assert '7' == config.get('log', 'interval')
    assert '8' == config.log_backups
    assert '8' == config.get('log', 'backups')


def test_install__Configurator__get_log_options__2(config, capsys):
    """It asks the user about logging and store his answers.

    Choosing `RotatingFileHandler` two more questions are asked.
    """
    with user_input(['RotatingFileHandler', '12345', '2'], config):
        config.get_log_options()
    assert ([
        u' Please choose Log-Handler:',
        u'    Details see http://docs.python.org/library/logging.html#'
        u'handler-objects',
        u' Log-Handler, choose between FileHandler, RotatingFileHandler, '
        u'TimedRotatingFileHandler: [FileHandler] ',
        u' Maximum file size before rotating in bytes: [1000] ',
        u' Number of log file backups: [5] ',
    ] == capsys.readouterr()[0].splitlines())
    assert 'RotatingFileHandler' == config.log_handler
    assert 'RotatingFileHandler' == config.get('log', 'handler')
    assert '12345' == config.log_max_bytes
    assert '12345' == config.get('log', 'max_size')
    assert '2' == config.log_backups
    assert '2' == config.get('log', 'backups')


def test_install__Configurator__get_log_options__3(config, capsys):
    """It asks the user about logging and store his answers.

    Choosing `FileHandler` no more questions are asked:
    """
    with user_input('FileHandler', config):
        config.get_log_options()
    assert ([
        u' Please choose Log-Handler:',
        u'    Details see http://docs.python.org/library/logging.html#'
        u'handler-objects',
        u' Log-Handler, choose between FileHandler, RotatingFileHandler, '
        u'TimedRotatingFileHandler: [FileHandler] ',
    ] == capsys.readouterr()[0].splitlines())
    assert 'FileHandler' == config.log_handler
    assert 'FileHandler' == config.get('log', 'handler')


def test_install__Configurator__print_additional_packages_intro__1(
        config, capsys):
    """It prints the introduction text for additional packages."""
    config.print_additional_packages_intro()
    assert ([
        u' When you need additional packages (e. g. import readers)',
        u' enter the package names here separated by a newline.',
        u' When you are done enter an empty line.',
        u' Known packages:',
        u'   icemac.ab.importer -- Import of CSV files',
        u'   icemac.ab.importxls -- Import of XLS (Excel) files',
        u'   icemac.ab.calendar -- Calendar using persons in address book',
    ] == capsys.readouterr()[0].splitlines())


def test_install__Configurator__get_additional_packages__1(config, capsys):
    """It asks for the names of additional packages.

    Those packages are later added as runtime and test dependenies. They need a
    `configure.zcml` as they get integrated using that file.
    """
    with user_input(['icemac.ab.importxls', 'icemac.ab.importcsv'], config):
        config.get_additional_packages()
    assert ([
        u' Package 1: [] ',
        u' Package 2: [] ',
        u' Package 3: [] ',
    ] == capsys.readouterr()[0].splitlines())
    assert ['icemac.ab.importxls', 'icemac.ab.importcsv'] == config.packages
    assert 'icemac.ab.importxls' == config.get('packages', 'package_1')
    assert 'icemac.ab.importcsv' == config.get('packages', 'package_2')
    assert '' == config.get('packages', 'package_3')


def test_install__Configurator__get_migration_options__1(config, capsys):
    """It ask the user about the migration of existing content.

    It stores the values. When the user decides no to do the migration no
    additional questions are asked.
    """
    with user_input('no', config):
        config.get_migration_options()
    assert (u'  Migrate old address book content to new instance: [no] \n' ==
            capsys.readouterr()[0])
    assert 'no' == config.do_migration
    assert 'no' == config.get('migration', 'do_migration')


def test_install__Configurator__get_migration_options__2(config, capsys):
    """It ask the user about the migration of existing content.

    When the user chooses to migrate he is asked some additional questions.
    """
    with user_input(['yes', 'no', 'yes'], config):
        config.get_migration_options()
    assert ([
        u'  Migrate old address book content to new instance: [no] ',
        u'The old address book instance must not run during migration.',
        u'When it runs as demon process the migration script can stop it '
        u'otherwise you have to stop it manually.',
        u' Old instance is running as a demon process: [no] ',
        u' New instance should be started as a demon process after migration: '
        u'[no] ',
    ] == capsys.readouterr()[0].splitlines())
    assert 'yes' == config.do_migration
    assert 'yes' == config.get('migration', 'do_migration')
    assert 'no' == config.stop_server
    assert 'no' == config.get('migration', 'stop_server')
    assert 'yes' == config.start_server
    assert 'yes' == config.get('migration', 'start_server')


def test_install__Configurator__create_admin_zcml__1(config, capsys, basedir):
    """It creates the `admin.zcml` file.

    This file which contains the password of the global administrator.
    """
    config.admin_login = 'root'
    config.admin_passwd = 'keins'
    config.create_admin_zcml()
    assert 'creating admin.zcml ...\n' == capsys.readouterr()[0]
    assert [
        '<configure xmlns="http://namespaces.zope.org/zope">',
        '  <principal',
        '    id="icemac.addressbook.global.Administrator"',
        '    title="global administrator"',
        '    login="root"',
        '    password_manager="Plain Text"',
        '    password="keins" />',
        '  <grant',
        '    role="icemac.addressbook.global.Administrator"',
        '    principal="icemac.addressbook.global.Administrator" />',
        '  <grant',
        '    permission="zope.ManageContent"',
        '    principal="icemac.addressbook.global.Administrator" />',
        '</configure>',
    ] == basedir.join('admin.zcml').read().splitlines()


def test_install__Configurator__create_buildout_cfg__1(
        config, capsys, basedir):
    """It creates the `buildout.cfg` file.

    This file contains the user configurations for the server. The contents of
    the file depend a bit on the chosen log handler.
    """
    config.eggs_dir = '/var/lib/eggs'
    config.host = 'server.local'
    config.port = '8081'
    config.username = ''
    config.log_handler = 'FileHandler'
    config.packages = ['icemac.ab.reporting', 'icemac.ab.relations']
    config.create_buildout_cfg()
    assert 'creating buildout.cfg ...\n' == capsys.readouterr()[0]
    assert [
        '[buildout]',
        'extends = profiles/prod.cfg',
        'newest = true',
        'allow-picked-versions = true',
        'eggs-directory = /var/lib/eggs',
        '',
        '[deploy.ini]',
        'host = server.local',
        'port = 8081',
        'log-handler = FileHandler',
        "log-handler-args = 'a'",
        '',
        '[site.zcml]',
        'permissions_zcml = <include package="icemac.ab.reporting" />',
        '\t<include package="icemac.ab.relations" />',
        '',
        '[app]',
        'eggs += icemac.ab.reporting',
        '      icemac.ab.relations',
        '',
        '[test]',
        'eggs += icemac.ab.reporting',
        '      icemac.ab.relations',
        ''
    ] == basedir.join('buildout.cfg').read().splitlines()


def test_install__Configurator__create_buildout_cfg__2(
        config, capsys, basedir):
    """It creates the `buildout.cfg` file.

    Using different log handler results in different ``log-handler-args``.
    """
    config.log_handler = 'RotatingFileHandler'
    config.log_max_bytes = '200000'
    config.log_backups = '10'
    config.eggs_dir = 'eggs'
    config.host = 'host'
    config.port = 'port'
    config.username = ''
    config.packages = []
    config.create_buildout_cfg()
    assert 'creating buildout.cfg ...\n' == capsys.readouterr()[0]
    assert ("log-handler-args = 'a', 200000, 10" in
            basedir.join('buildout.cfg').read())


def test_install__Configurator__create_buildout_cfg__3(
        config, capsys, basedir):
    """It creates the `buildout.cfg` file.

    Using `TimedRotatingFileHandler` results in different ``log-handler-args``.
    """
    config.log_handler = 'TimedRotatingFileHandler'
    config.log_when = 'W'
    config.log_interval = '2'
    config.log_backups = '1'
    config.eggs_dir = 'eggs'
    config.host = 'host'
    config.port = 'port'
    config.username = ''
    config.packages = []
    config.create_buildout_cfg()
    assert 'creating buildout.cfg ...\n' == capsys.readouterr()[0]
    assert "log-handler-args = 'W', 2, 1" in basedir.join(
        'buildout.cfg').read()


def test_install__Configurator__create_buildout_cfg__4(
        config, capsys, basedir):
    """It creates the `buildout.cfg` file.

    When the username is not empty a zdaemon.conf section is added.
    """
    config.username = 'mac'
    config.log_handler = 'FileHandler'
    config.eggs_dir = 'eggs'
    config.host = 'host'
    config.port = 'port'
    config.packages = []
    config.create_buildout_cfg()
    assert 'creating buildout.cfg ...\n' == capsys.readouterr()[0]
    assert ("[zdaemon.conf]\nuser = user mac" in
            basedir.join('buildout.cfg').read())


def test_install__Configurator__store__1(config, capsys, basedir):
    """It stores the configuration values in a file named `install.user.ini`"""
    config.store()
    assert 'saving config ...\n' == capsys.readouterr()[0]
    assert [
        '[install]',
        'eggs_dir = py-eggs',
        '',
        '[admin]',
        'login = me',
        'password = secret',
        '',
        '[server]',
        'host = my.computer.local',
        'port = 13090',
        'username = ',
        '',
        '[log]',
        'handler = FileHandler',
        'max_size = 1000',
        'when = midnight',
        'interval = 1',
        'backups = 5',
        '',
        '[packages]',
        '',
        '[migration]',
        'do_migration = no',
        'stop_server = no',
        'start_server = no',
        'old_instance = ',
        '',
    ] == basedir.join('install.user.ini').read().splitlines()


def test_install__Configurator____call____1(config, capsys, basedir):
    """It runs the complete configuration.

    To ease testing changes we use only default values.
    """
    with user_input('', config):
        config()
    assert [
        'Welcome to icemac.addressbook installation',
        '',
        'Hint: to use the default value (the one in [brackets]),  enter no '
        'value.',
        '',
        ' Directory to store python eggs: [py-eggs] ',
        ' Log-in name for the administrator: [me] ',
        ' Password for the administrator: [secret] ',
        ' Hostname: [my.computer.local] ',
        ' Port number: [13090] ',
        ' Username whether process should run as different user otherwise '
        'emtpy: [] ',
        ' Please choose Log-Handler:',
        '    Details see http://docs.python.org/library/logging.html#'
        'handler-objects',
        ' Log-Handler, choose between FileHandler, RotatingFileHandler, '
        'TimedRotatingFileHandler: [FileHandler] ',
        ' When you need additional packages (e. g. import readers)',
        ' enter the package names here separated by a newline.',
        ' When you are done enter an empty line.',
        ' Known packages:',
        '   icemac.ab.importer -- Import of CSV files',
        '   icemac.ab.importxls -- Import of XLS (Excel) files',
        '   icemac.ab.calendar -- Calendar using persons in address book',
        ' Package 1: [] ',
        '  Migrate old address book content to new instance: [no] ',
        'creating admin.zcml ...',
        'creating buildout.cfg ...',
        'saving config ...',
    ] == capsys.readouterr()[0].splitlines()
    assert [
        '[buildout]',
        'extends = profiles/prod.cfg',
        'newest = true',
        'allow-picked-versions = true',
        'eggs-directory = py-eggs',
        '',
        '[deploy.ini]',
        'host = my.computer.local',
        'port = 13090',
        'log-handler = FileHandler',
        "log-handler-args = 'a'",
        '',
    ] == basedir.join('buildout.cfg').read().splitlines()
    assert [
        '[install]',
        'eggs_dir = py-eggs',
        '',
        '[admin]',
        'login = me',
        'password = secret',
        '',
        '[server]',
        'host = my.computer.local',
        'port = 13090',
        'username = ',
        '',
        '[log]',
        'handler = FileHandler',
        'max_size = 1000',
        'when = midnight',
        'interval = 1',
        'backups = 5',
        '',
        '[packages]',
        'package_1 = ',
        '',
        '[migration]',
        'do_migration = no',
        'stop_server = no',
        'start_server = no',
        'old_instance = ',
        '',
    ] == basedir.join('install.user.ini').read().splitlines()
    assert [
        '<configure xmlns="http://namespaces.zope.org/zope">',
        '  <principal',
        '    id="icemac.addressbook.global.Administrator"',
        '    title="global administrator"',
        '    login="me"',
        '    password_manager="Plain Text"',
        '    password="secret" />',
        '  <grant',
        '    role="icemac.addressbook.global.Administrator"',
        '    principal="icemac.addressbook.global.Administrator" />',
        '  <grant',
        '    permission="zope.ManageContent"',
        '    principal="icemac.addressbook.global.Administrator" />',
        '</configure>',
    ] == basedir.join('admin.zcml').read().splitlines()



from icemac.addressbook.browser.search.result.handler.manager import (
    SearchResultHandler)


def makeSRHandler(viewName):
    """Create a `SearchResultHandler` with the specified `viewName`."""
    handler = SearchResultHandler(None, None, None, None)
    handler.viewName = viewName
    return handler


def test_manager__SearchResultHandler___eq___1():
    """SearchResultHandler instances are equal when `viewName` is equal."""
    assert makeSRHandler('@@asdf.html') == makeSRHandler('@@asdf.html')


def test_manager__SearchResultHandler___eq___2():
    """SearchResultHandler instances are not equal with unequal `viewName`."""
    # There is no __neq__ implemented!
    assert not(makeSRHandler('@@foo.html') == makeSRHandler('@@bar.html'))


def test_manager__SearchResultHandler___eq___3():
    """A SearchResultHandler instance is not equal to anything else."""
    # There is no __neq__ implemented!
    assert not(makeSRHandler(None) == object())



import zope.interface


class IDontReuseNames(zope.interface.Interface):
    """Marker interface for container those names are never reused.

    Even when contained objects get deleted their names are _not_ reused.
    """


class INameSuffix(zope.interface.Interface):
    """Suffix to make name in container unique."""

    def __iadd__(value):
        """Increment the suffix by `value`."""

    def __unicode__():
        """Convert suffix to unicode."""



from __future__ import unicode_literals
import pytest


@pytest.fixture(scope='function')
def md_menu(address_book, browser, sitemenu):
    """Fixture to test the master data menu."""
    browser.login('mgr')
    return sitemenu(
        browser, 3, 'Master data', browser.MASTER_DATA_URL)


def test_menu__master_data_menu__1(md_menu):
    """Asserting that the menu with the index 3 is `Master Data`."""
    md_menu.assert_correct_menu_item_is_tested()


def test_menu__master_data_menu__2(md_menu):
    """The master data menu item is selected on the master data overview."""
    assert md_menu.item_selected(md_menu.menu_item_URL)


def test_menu__master_data_menu__3(md_menu):
    """The master data menu item is not selected on the person list."""
    assert not md_menu.item_selected(md_menu.browser.PERSONS_LIST_URL)


def test_menu__master_data_menu__4(md_menu):
    """The master data menu item is selected on address book edit."""
    assert md_menu.item_selected(md_menu.browser.ADDRESS_BOOK_EDIT_URL)


def test_menu__master_data_menu__5(md_menu):
    """The master data menu item is selected on the keywords add form."""
    assert md_menu.item_selected(md_menu.browser.KEYWORD_ADD_URL)


def test_menu__master_data_menu__6(md_menu):
    """The master data menu item is selected on the principals list."""
    assert md_menu.item_selected(md_menu.browser.PRINCIPALS_LIST_URL)


def test_menu__master_data_menu__7(md_menu):
    """The master data menu item is selected on user field add form."""
    assert md_menu.item_selected(md_menu.browser.ENTITY_PERSON_ADD_FIELD_URL)



# -*- coding: utf-8 -*-
from js.jquery_timepicker_addon import timepicker_locales, timepicker
from js.jqueryui import base as jqueryui_css
from js.jqueryui import ui_datepicker_locales, ui_selectable
import grokcore.component as grok
import icemac.addressbook.browser.interfaces
import icemac.addressbook.interfaces
import icemac.addressbook.preferences.utils
import json
import pytz
import z3c.form.browser.select
import z3c.form.browser.text
import z3c.form.converter
import z3c.form.hint
import z3c.form.interfaces
import z3c.form.widget
import zope.interface
import zope.schema.interfaces


class DatetimeDataConverter(z3c.form.converter.DatetimeDataConverter):
    """Converter for z3c.form displaying localized date times.

    Converts stored datetimes to timezone of request.
    Converts datetimes from form to store with time zone of request.

    """

    @property
    def time_zone(self):
        """User selected time zone."""
        return icemac.addressbook.preferences.utils.get_time_zone()

    def toWidgetValue(self, value):
        """Convert to time zone user has selected."""
        if value is not self.field.missing_value and value.tzinfo is not None:
            value = value.astimezone(self.time_zone)
        return super(DatetimeDataConverter, self).toWidgetValue(value)

    def toFieldValue(self, value):
        value = super(DatetimeDataConverter, self).toFieldValue(value)
        if value is None:
            return value
        return self.time_zone.localize(value).astimezone(pytz.UTC)


# Map zope.i18n binary pattern to jquery pattern:
# Sources:
# * http://api.jqueryui.com/datepicker
# * http://trentrichardson.com/examples/timepicker/#tp-formatting
# * https://github.com/zopefoundation/zope.i18n/blob/
#           b8fff2676a7575bf2d3248c562f778776e0603ce/src/zope/i18n/interfaces/
#           __init__.py#L377
pattern_mapping = {
    'date': {
        ('y', 2): 'y',
        ('y', 4): 'yy',
        ('M', 1): 'm',
        ('M', 2): 'mm',
        ('M', 3): 'M',
        ('M', 4): 'MM',
        ('D', 1): 'o',
        ('D', 2): 'o',
        ('D', 3): 'oo',
        ('E', 3): 'D',
        ('E', 4): 'DD',
    },
    'time': {
        ('H', 1): 'h',
        ('H', 2): 'HH',
        ('h', 1): 'h',
        ('S', 1): 'l',
        ('a', 1): 'TT',
        ('a', 2): 'TT',
    }
}
pattern_mapping['datetime'] = pattern_mapping['date'].copy()
pattern_mapping['datetime'].update(pattern_mapping['time'])
# Mappings of jquery and zope.i18n do not exactly match:
pattern_mapping['datetime'][('h', 1)] = 'hh'


def zope_i18n_pattern_to_jquery_pattern(kind, pattern):
    """Convert a zope.i18n datetime pattern to one of jquery."""
    # bin_pattern is a list of tuples (char, count) or chars
    bin_pattern = zope.i18n.format.parseDateTimePattern(pattern)
    # pattern_mapping contains only the differences, so by default produce
    # char * count (or the char if it is no tuple):
    pattern = ''.join(
        [pattern_mapping[kind].get(
            x, x[0] * x[1] if isinstance(x, tuple) else x)
         for x in bin_pattern])
    return pattern


class DatetimeWidgetBase(z3c.form.browser.text.TextWidget):
    """Base class widget to enter date and time using JavaScript picker."""

    klass = NotImplemented
    picker_js_func_name = NotImplemented
    type = NotImplemented

    def get_parameters(self):
        converter = z3c.form.interfaces.IDataConverter(self)
        params = {}
        if 'date' in self.picker_js_func_name:
            params.update(
                {'dateFormat': self.get_format(converter, 'date', self.type),
                 'changeMonth': True,
                 'changeYear': True,
                 'yearRange': '-100:+10',
                 'firstDay': 0,
                 'numberOfMonths': 3,
                 'stepMonths': 3})
        if 'time' in self.picker_js_func_name:
            params.update(
                {'timeFormat': self.get_format(converter, 'time', self.type),
                 'hourGrid': 4,
                 'hourMin': 6,
                 'hourMax': 22,
                 'minuteGrid': 15,
                 'stepMinute': 15})
        return params

    def get_format(self, converter, type, for_):
        """Get the jQuery format string from a z3c.form converter.

        converter ... z3c.form data converter instance
        type ... 'date' or 'time'
        for_ ... 'date', 'time', or 'datetime'

        """
        getFormatter = self.request.locale.dates.getFormatter
        return zope_i18n_pattern_to_jquery_pattern(
            for_, getFormatter(type, converter.length).getPattern())

    def update(self):
        super(DatetimeWidgetBase, self).update()
        locale_name = self.request.locale.id.language
        jqueryui_css.need()
        timepicker.need()
        if locale_name in timepicker_locales:
            timepicker_locales[locale_name].need()
        if locale_name in ui_datepicker_locales:
            ui_datepicker_locales[locale_name].need()

    def javascript(self):
        return "jQuery('#%s').%s(%s);" % (
            self.id, self.picker_js_func_name,
            json.dumps(self.get_parameters()))


@zope.interface.implementer(
    icemac.addressbook.browser.interfaces.IDatetimeWidget)
class DatetimeWidget(DatetimeWidgetBase):
    """Widget to enter date and time using JavaScript picker."""

    klass = 'datetime-widget'
    picker_js_func_name = 'datetimepicker'
    type = 'datetime'


@grok.adapter(
    zope.schema.interfaces.IDatetime,
    icemac.addressbook.browser.interfaces.IAddressBookLayer)
@grok.implementer(z3c.form.interfaces.IFieldWidget)
def DatetimeFieldWidget(field, request):
    """Factory for DatetimeWidget."""
    return z3c.form.widget.FieldWidget(field, DatetimeWidget(request))


@zope.interface.implementer(
    icemac.addressbook.browser.interfaces.IDateWidget)
class DateWidget(DatetimeWidgetBase):
    """Widget to enter date using a JavaScript picker."""

    klass = 'date-widget'
    picker_js_func_name = 'datepicker'
    type = 'date'


@grok.adapter(
    zope.schema.interfaces.IDate,
    icemac.addressbook.browser.interfaces.IAddressBookLayer)
@grok.implementer(z3c.form.interfaces.IFieldWidget)
def DateFieldWidget(field, request):
    """Factory for DateWidget."""
    return z3c.form.widget.FieldWidget(field, DateWidget(request))


@zope.interface.implementer(
    icemac.addressbook.browser.interfaces.ITimeWidget)
class TimeWidget(DatetimeWidgetBase):
    """Widget to enter time using JavaScript picker."""

    klass = 'time-widget'
    picker_js_func_name = 'timepicker'
    type = 'time'


@grok.adapter(
    zope.schema.interfaces.ITime,
    icemac.addressbook.browser.interfaces.IAddressBookLayer)
@grok.implementer(z3c.form.interfaces.IFieldWidget)
def TimeFieldWidget(field, request):
    """Factory for TimeWidget."""
    return z3c.form.widget.FieldWidget(field, TimeWidget(request))


class DateDataConverter(z3c.form.converter.DateDataConverter):
    """Special date converter which does not have a year 2k problem."""

    length = 'medium'


class ImageSelectWidget(z3c.form.browser.select.SelectWidget):
    """Select widget displays images as selectables."""

    zope.interface.implements(
        icemac.addressbook.browser.interfaces.IImageSelectWidget)

    def update(self):
        super(ImageSelectWidget, self).update()
        jqueryui_css.need()
        ui_selectable.need()


@grok.adapter(zope.schema.interfaces.IChoice,
              icemac.addressbook.interfaces.IImageSource,
              icemac.addressbook.browser.interfaces.IAddressBookLayer)
@grok.implementer(z3c.form.interfaces.IFieldWidget)
def SelectFieldWidget(field, source, request):
    """`IFieldWidget` factory for SelectWidget."""
    return z3c.form.widget.FieldWidget(field, ImageSelectWidget(request))


class FieldDescriptionAsHint(z3c.form.hint.FieldDescriptionAsHint):
    """Get description or adapter registered as `hint`."""

    def get(self):
        computed_value = zope.component.queryMultiAdapter(
            (self.context, self.request, self.form, self.field, self.widget),
            name='hint')
        if computed_value is None:
            desc = super(FieldDescriptionAsHint, self).get()
            if desc and set(desc).intersection(set('\r\n')):
                desc = desc.replace('\r', ' ').replace('\n', ' ')
        else:
            desc = computed_value.get()
        return desc



from mock import patch


def test_icalendar__iCalendar__1(search_data, browser):
    """The view returns an iCalendar file containing the birthday."""
    browser.login('visitor')
    browser.open(browser.PERSONS_LIST_URL)
    browser.getLink('Liese').click()
    browser.getControl('Export').click()
    with patch('pkg_resources.get_distribution') as get_distribution:
        get_distribution().version = '1.2.3'
        browser.getLink('iCalendar').click()
    assert ['BEGIN:VCALENDAR',
            'VERSION:2.0',
            'PRODID:-//icemac.addressbook//1.2.3//EN',
            'CALSCALE:GREGORIAN',
            'BEGIN:VEVENT',
            'SUMMARY:Tester\\, Liese (*1976)',
            'DTSTART;VALUE=DATE:19761115',
            'DTEND;VALUE=DATE:19761116',
            'UID:http://localhost/ab/Person-5/@@iCalendar',
            'RRULE:FREQ=YEARLY',
            'END:VEVENT',
            'END:VCALENDAR',
            ''] == browser.contents.split('\r\n')



import pytest
from icemac.addressbook.catalog.serializer import FieldNoSerializer
from icemac.addressbook.catalog.serializer import FieldSerializer
import gocept.country.db
import zope.schema


class SerializableObject(object):
    """Test object for serialization tests."""

    text = u'asdf'
    number1 = 2411
    number2 = 1234.6789
    bytes = 'foobar'


def callSerializer(field_class_name, field_name, serializer):
    """Call a serializer on a specified field type."""
    field = getattr(zope.schema, field_class_name)(title=u'')
    field.__name__ = field_name
    result = serializer(field, SerializableObject())
    assert isinstance(result, unicode)
    return result


def test_serializer__FieldSerializer__1():
    """`FieldSerializer` converts TextLine to unicode."""
    assert u'asdf' == callSerializer('TextLine', 'text', FieldSerializer)


def test_serializer__FieldSerializer__2():
    """`FieldSerializer` converts Text to unicode."""
    assert u'asdf' == callSerializer('Text', 'text', FieldSerializer)


def test_serializer__FieldSerializer__3():
    """`FieldSerializer` converts Int to unicode."""
    assert u'2411' == callSerializer('Int', 'number1', FieldSerializer)


def test_serializer__FieldSerializer__4():
    """`FieldSerializer` converts Float to unicode."""
    assert u'1234.6789' == callSerializer('Float', 'number2', FieldSerializer)


def test_serializer__FieldNoSerializer__1():
    """`FieldSerializer` converts Float to unicode."""
    assert u'' == callSerializer('Bytes', 'bytes', FieldNoSerializer)


@pytest.mark.xfail
def test_serializer__ObjectSerializer__1(
        address_book, PersonFactory, PostalAddressFactory):
    """`ObjectSerializer` serializes all field values."""
    person = PersonFactory(address_book, u'Tester')
    PostalAddressFactory(
        person, city=u'Dunkelhausen', country=gocept.country.db.Country('CH'))
    # XXX test not yet complete
    assert 0
